/** @file cf_io.c
  * @brief This file contains CF bus driver.
  * 
  * Copyright  Marvell International Ltd. and/or its affiliates, 2003-2006
  * 
  * This software file (the "File") is owned and distributed by 
  * Marvell International Ltd. and/or its affiliates ("Marvell") 
  * under the terms and conditions of the General Public License 
  * Version 2, June 1991 (the "GPL License"), a copy of which is 
  * available along with the File in the license.txt file or by
  * writing to the Free Software Foundation, Inc., 59 Temple Place, 
  * Suite 330, Boston, MA 02111-1307 or on the worldwide web at 
  * http://www.gnu.org/licenses/gpl.txt. 
  * 
  * THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, 
  * AND THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR 
  * A PARTICULAR PURPOSE ARE EXPRESSLY DISCLAIMED.  The GPL License 
  * provides additional details about this warranty disclaimer.
  *	
  */

/********************************************************
Change log:
	10/04/05: Add Doxygen format comments
	1/12/06: Add Kernel 2.6 support
	
********************************************************/

#include  "cfio_io.h"

/********************************************************
		Local Variables
********************************************************/

#define __NO_VERSION__

#ifdef LINUX_2_4
#define CS_ERROR(h,f,r) { \
        error_info_t  err = {f,r}; \
	pcmcia_report_error(h,&err); \
      }
#else
#define CS_ERROR(h,f,r) cs_error(h,f,r)
#endif
#define CF_IRQMASK	  0xdeb8
#define INT_MODULE_PARM(n, v) static int n = v; MODULE_PARM(n, "i")
#define IRQINFO2_LEN	  4
#define CF_VERSION	  0x0326

dev_info_t cfio_dev_info = "mcf25";

dev_link_t *dev_list 	 = NULL;

MODULE_PARM(interrupt_steer, "1-4i");

INT_MODULE_PARM(irq_mask, CF_IRQMASK);

/* Module Variables */
static int interrupt_steer[IRQINFO2_LEN] = { -1 };

typedef struct _if_pcmcia_info_t {
  dev_link_t link;
  dev_node_t node;
  int stop;
  struct bus_operations *bus;
  struct net_device *eth_dev;
} if_pcmcia_info_t;

/********************************************************
		Global Variables
********************************************************/

struct cf_card_rec cardp;

/********************************************************
		Local Functions
********************************************************/

/** 
 *  @brief release allocated CF bus driver resource
 *  @param arg       pointer to dev_link_t
 *  @return 	     N/A
 */
static void cf_release(ulong arg)
{
  dev_link_t *link = (dev_link_t *) arg;

  link->dev = NULL;

  link->conf.Vcc = 0;
  pcmcia_release_configuration(link->handle);
  if (link->io.NumPorts1)
    pcmcia_release_io(link->handle, &link->io);

  if (link->irq.AssignedIRQ)
    pcmcia_release_irq(link->handle, &link->irq);

  link->state &= ~DEV_CONFIG;

} /* cf_release */



/** 
 *  @brief detach CF BUS driver 
 *  @param link      pointer to dev_link_t
 *  @return 	     N/A
 */

static void cf_detach(dev_link_t * link)
{
  dev_link_t **p;

  for (p = &dev_list; *p; p = &(*p)->next)
    if (*p == link)
      break;

  if (*p == NULL)
    return;

#ifdef LINUX_2_4
  del_timer_sync(&link->release);
#endif

  if (((if_pcmcia_info_t *) link->priv)->eth_dev) {
    printk("Before calling wlan_remove function\n");
    cardp.remove(&cardp);
    printk("After calling wlan_remove function\n");
  }

  if (link->state & DEV_CONFIG) {
    cf_release((u32) link);
  }

  ((if_pcmcia_info_t *) link->priv)->eth_dev = NULL;

  if (link->handle) {
#ifdef LINUX_2_4
    pcmcia_reset_card(link->handle, NULL);
#endif
    pcmcia_deregister_client(link->handle);
  }

  *p = link->next;
  /* This points to the parent if_pcmcia_info_t struct */
  if (link->priv)
    kfree(link->priv);
} /* cf_detach */

/** 
 *  @brief CF BUS driver Configuration
 *  @param link      pointer to dev_link_t
 *  @return 	     N/A
 */
static void cf_config(dev_link_t * link)
{
  client_handle_t handle = link->handle;
  if_pcmcia_info_t *dev = link->priv;
  tuple_t tuple;
  cisparse_t parse;
  cistpl_cftable_entry_t *cfg = &parse.cftable_entry;
#define BUF_LEN 64
  u8 buf[BUF_LEN];
  config_info_t conf;

  tuple.DesiredTuple = CISTPL_CONFIG;
  tuple.Attributes = 0;
  tuple.TupleData = buf;
  tuple.TupleDataMax = sizeof(buf);
  tuple.TupleOffset = 0;

  if (pcmcia_get_first_tuple(handle, &tuple))
    goto onerror;
  if (pcmcia_get_tuple_data(handle, &tuple))
    goto onerror;
  if (pcmcia_parse_tuple(handle, &tuple, &parse))
    goto onerror;

  link->conf.ConfigBase = parse.config.base;
  link->conf.Present = parse.config.rmask[0];

  link->state |= DEV_CONFIG;

  if (pcmcia_get_configuration_info(handle, &conf))
    goto onerror;

  link->conf.Vcc = conf.Vcc;

  /*
     The Configuration table consists of a series of configuration table
     entry tuples. Each entry consists of up to seven data structures that
     describe operational characteristsics of the PC card.
   */

  tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;

  if (pcmcia_get_first_tuple(handle, &tuple))
    goto onerror;

  if (pcmcia_get_tuple_data(handle, &tuple) != CS_SUCCESS)
    goto onerror;

  if (pcmcia_parse_tuple(handle, &tuple, &parse) != CS_SUCCESS)
    goto onerror;


  link->conf.ConfigIndex = cfg->index;

  /* Interrupt request description */
  if (cfg->irq.IRQInfo1)
    link->conf.Attributes |= CONF_ENABLE_IRQ;

  /* IO Address space description */
  link->io.NumPorts1 = link->io.NumPorts2 = 0;
  if ((cfg->io.nwin > 0)) {
    cistpl_io_t *io;

    if (cfg->io.nwin)
      io = &cfg->io;

    link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;

    if (!(io->flags & CISTPL_IO_8BIT))
      link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;

    if (!(io->flags & CISTPL_IO_16BIT))
      link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;

    link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;

    link->io.BasePort1 = io->win[0].base;
    link->io.NumPorts1 = io->win[0].len;

    if (io->nwin > 1) {
      link->io.Attributes2 = link->io.Attributes1;
      link->io.BasePort2 = io->win[1].base;
      link->io.NumPorts2 = io->win[1].len;
    }

    if (pcmcia_request_io(link->handle, &link->io)
        != CS_SUCCESS) {
      pcmcia_release_io(link->handle, &link->io);
      printk("Request IO Error !!\n");
      goto onerror;
    }
  }

  if (link->conf.Attributes & CONF_ENABLE_IRQ)
    if (pcmcia_request_irq(link->handle, &link->irq))
      goto onerror;

  if (pcmcia_request_configuration(link->handle, &link->conf))
    goto onerror;

  cardp.irq = link->irq.AssignedIRQ;
  cardp.port = link->io.BasePort1;
  printk("BasePort1=0x%x, AssignedIRQ=%d\n",
         link->io.BasePort1, link->irq.AssignedIRQ);

  if (!(cardp.add(&cardp))) {
    printk("Call to cardp.add failed\n");
    goto onerror;
  }

  printk("After calling wlan_add_card function\n");

  ((if_pcmcia_info_t *) link->priv)->eth_dev = cardp.eth_dev;

  if (!((if_pcmcia_info_t *) link->priv)->eth_dev)
    goto onerror;

  strcpy(dev->node.dev_name, cardp.eth_dev->name);
  dev->node.major = dev->node.minor = 0;
  link->dev = &dev->node;

  link->state &= ~DEV_CONFIG_PENDING;
  return;

onerror:
  printk("card configuration failed...calling cf_release function\n");
  cf_release((u32) link);
  link->state &= ~DEV_CONFIG_PENDING;
  cardp.flag = 1;

} /* cf_config */


/** 
 *  @brief CF BUS driver Event handler
 *  @param event     event id
 *  @param priority  event priority
 *  @param args      pointer to event_callback_args_t
 *  @return 	     0 
 */
static int cf_event(event_t event, int priority, event_callback_args_t *args)
{
  dev_link_t *link = args->client_data;
  if_pcmcia_info_t *dev = link->priv;

  switch (event) {
  case CS_EVENT_CARD_INSERTION:
    link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
#ifdef LINUX_2_4
    dev->bus = args->bus;
#endif
    cf_config(link);
    break;

  case CS_EVENT_CARD_REMOVAL:
    link->state &= ~DEV_PRESENT;
    printk("card removal event detected\n");
    if (link->state & DEV_CONFIG) {
      ((if_pcmcia_info_t *) link->priv)->stop = 1;
      printk("Before calling release function\n");
#ifdef LINUX_2_4
#define REMOVE_TIMEOUT		(HZ/20)
      mod_timer(&link->release, jiffies + REMOVE_TIMEOUT);
#else
      cf_detach(link);
#endif
      printk("After calling release function\n");
    }
    break;
  case CS_EVENT_CARD_RESET:
    if (link->state & DEV_CONFIG) {
      pcmcia_request_configuration(link->handle, &link->conf);
    }
    dev->stop = 0;
    break;
  }

  return 0;
}       /* cf_event */

/** 
 *  @brief attach CF BUS driver 
 *  @return 	     pointer to dev_link_t
 */

static dev_link_t *cf_attach(void)
{
  u8 i; 
  int status;
  if_pcmcia_info_t *ifinfo;
  dev_link_t *link;
  client_reg_t client_reg;

  printk("Entering cf_attach()\n");

  /* Allocate space for PCMCIA information */
  if (!(ifinfo = kmalloc(sizeof(if_pcmcia_info_t), GFP_KERNEL))) {
    return NULL;
  }

  memset(ifinfo, 0, sizeof(if_pcmcia_info_t));
  link = &ifinfo->link;
  link->priv = ifinfo;

#ifdef LINUX_2_4
  init_timer(&link->release);
  link->release.function = &cf_release;
  link->release.data = (ulong) link;
#endif

  link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
  link->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
  
  /* Is the IRQ specified by the user? */
  if (interrupt_steer[0] == -1) {
    /* No, find the required IRQ */
    link->irq.IRQInfo2 = irq_mask;
  } else {
    /* Yes, create the mask based on user specified IRQ's */
    for (i = 0; i < IRQINFO2_LEN; i++) {
      link->irq.IRQInfo2 |= 1 << interrupt_steer[i];
    }
  }

  /* Handler will be installed later */
  link->irq.Handler = NULL;

  link->conf.Attributes = 0;
#define VCC_VALUE	50
  link->conf.Vcc = VCC_VALUE;
  link->conf.Vpp1 = 0;
  link->conf.Vpp2 = 0;
  link->conf.IntType = INT_MEMORY_AND_IO;
  link->conf.ConfigIndex = 1;
  link->conf.Present = PRESENT_OPTION;

  link->next = dev_list;
  dev_list = link;
  client_reg.dev_info = &cfio_dev_info;
  client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
  client_reg.EventMask = CS_EVENT_CARD_INSERTION |
      CS_EVENT_CARD_REMOVAL | CS_EVENT_CARD_RESET;
  client_reg.event_handler = &cf_event;
  client_reg.event_callback_args.client_data = link;
  client_reg.Version = CF_VERSION;

  printk("Before registering the client\n");

  if ((status = pcmcia_register_client(&link->handle,
           &client_reg)) != CS_SUCCESS) {
    printk("Registering the client failed\n");
    CS_ERROR(link->handle, RegisterClient, status);
    cf_detach(link);
    return NULL;
  }

  printk("Leaving cf_attach()\n");

  return link;
}       /* cf_attach */


#ifndef LINUX_2_4
static struct pcmcia_driver cfio_cs_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "mcf25"
	},
	.attach		= cf_attach,
	.detach		= cf_detach,
	.event		= cf_event,
};
#endif

/********************************************************
		Global Functions
********************************************************/

/** 
 *  @brief register CF bus driver with OS and CardService
 *  @param add       add callback function
 *  @param remove    remove callback function
 *  @param arg       pointer to callback argument
 *  @return 	     pointer to card data
 */
u32 *register_cf_driver(cf_notifier_fn_add add, 
			cf_notifier_fn_remove remove, void *arg)
{
#ifdef LINUX_2_4
  servinfo_t serv;
#endif

  cardp.add = add;
  cardp.remove = remove;
  cardp.host_int_mask = 0;

#ifdef LINUX_2_4
  pcmcia_get_card_services_info(&serv);

  if (serv.Revision != CS_RELEASE_CODE) {
    return NULL;
  }
#endif

  printk("Before register driver\n");
#ifdef LINUX_2_4
  register_pccard_driver(&cfio_dev_info, &cf_attach, &cf_detach);
#else
  pcmcia_register_driver(&cfio_cs_driver);
#endif
  printk("After register driver\n");

  return (u32 *) &cardp;
}

/** 
 *  @brief unregister CF bus Driver
 *  @return 	     N/A
 */
void unregister_cf_driver( void )
{

#ifdef LINUX_2_4
  unregister_pccard_driver(&cfio_dev_info);
#else
  pcmcia_unregister_driver(&cfio_cs_driver);
#endif

  cf_detach(dev_list);

  while (dev_list != NULL) {
#ifdef LINUX_2_4
    del_timer(&dev_list->release);
#endif
    if (dev_list->state & DEV_CONFIG)
      cf_release((u32) dev_list);
  }
}


/** 
 *  @brief read CIS information
 *  @param priv      pointer to wlan_private
 *  @return 	     0
 */
s16 cfio_read_cfg_reg(void* priv)
{
  conf_reg_t reg;

  reg.Function  = 0;
  reg.Action    = CS_READ;
  reg.Offset    = 0;
  reg.Value     = 0;

  pcmcia_access_configuration_register(dev_list->handle, &reg);
  return 0;
}


/** 
 *  @brief module initialization procedure
 *  @return 	     N/A
 */
int cfio_init_module(void)
{
  return 0;
}

/** 
 *  @brief module cleanup
 *  @return 	     N/A
 */
void cfio_cleanup_module(void)
{
}

module_init(cfio_init_module);
module_exit(cfio_cleanup_module);

#ifdef LINUX_2_4
EXPORT_SYMBOL_NOVERS(cardp);
EXPORT_SYMBOL_NOVERS(cfio_dev_info);
EXPORT_SYMBOL_NOVERS(register_cf_driver);
EXPORT_SYMBOL_NOVERS(unregister_cf_driver);
EXPORT_SYMBOL_NOVERS(cfio_read_cfg_reg);
#else
EXPORT_SYMBOL(cardp);
EXPORT_SYMBOL(cfio_dev_info);
EXPORT_SYMBOL(register_cf_driver);
EXPORT_SYMBOL(unregister_cf_driver);
EXPORT_SYMBOL(cfio_read_cfg_reg);
#endif
