/*
 * Copyright  Marvell International Ltd. and/or its affiliates, 2003-2006
     This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef	__WLAN_THREAD_H_
#define	__WLAN_THREAD_H_

#ifdef LINUX_2_4
#include	<linux/smp_lock.h>
#else
#include	<linux/kthread.h>
#endif

#ifdef LINUX_2_4
#define WLAN_THREAD_STOPPED	0
#define WLAN_THREAD_RUNNING	1
#endif

typedef struct {
#ifdef LINUX_2_4
	struct completion	sync;
	struct tq_struct	taskQ;
#else
	struct task_struct	*task;
#endif
	wait_queue_head_t	waitQ;
	pid_t			pid;
	void			*priv;
#ifdef LINUX_2_4
	int			(*wlanfunc)(void *arg);
	int			state;
	char			name[16];
#endif
} wlan_thread;

static inline void wlan_activate_thread(wlan_thread *thr)
{
#ifdef LINUX_2_4
	/* Acquire kernel lock */
	lock_kernel();
#endif
	/** Record the thread pid */
	thr->pid = current->pid;
	
	/** Initialize the wait queue */
	init_waitqueue_head(&thr->waitQ);
#ifdef LINUX_2_4 
	/* Tell Signalling mechanism about the signals to be caught */
	siginitsetinv(&current->blocked, sigmask(SIGKILL) |
			sigmask(SIGINT)  | sigmask(SIGTERM));

	/* Set the state to running */
	thr->state = WLAN_THREAD_RUNNING;
  
	/* Setup the name of the thread */
	memcpy(current->comm, thr->name, sizeof(current->comm));
  
	/* Release the kernel lock */
	unlock_kernel();
  
	/* Unlock the process waiting for this activation */
	complete(&thr->sync);
#endif
}

static inline void wlan_deactivate_thread(wlan_thread *thr)
{
	ENTER();

#ifdef LINUX_2_4
	/* Acquire kernel lock */
	lock_kernel();

	/* Reset the pid */
	thr->pid = 0;
	
	/* Wake the process waiting for de-activation */
	complete(&thr->sync);
#else
	kthread_stop(thr->task);
#endif

	LEAVE();
}

#ifdef LINUX_2_4
static void wlan_register_thread(void *arg)
{
	wlan_thread	*thr = arg;
	
	/* Request the kernel to create a thread */
	kernel_thread(thr->wlanfunc, arg, 0);
}
#endif

static inline void wlan_create_thread(int (*wlanfunc)(void *), 
					wlan_thread *thr, char *name)
{
#ifdef LINUX_2_4
	/* Initialise wait mechanism */
	init_completion(&thr->sync);

	/* Setup the function to be run as the thread */
	thr->wlanfunc = wlanfunc;

	/* Record the thread name */
	memset(thr->name, 0, sizeof(thr->name));
	strncpy(thr->name, name, sizeof(thr->name)-1);
	
	/* Initialize the task queue */
	thr->taskQ.sync = 0;
	INIT_LIST_HEAD(&thr->taskQ.list);
	thr->taskQ.routine = wlan_register_thread;
	thr->taskQ.data = thr;
  
	/* Start the thread running */
	schedule_task(&thr->taskQ);

	/* Wait for thread to complete activation */
	wait_for_completion(&thr->sync);
#else
	thr->task = kthread_run(wlanfunc, thr, "%s", name);
#endif
}

static inline int wlan_terminate_thread(wlan_thread *thr)
{
	ENTER();

	/* Check if the thread is active or not */
	if (!thr->pid) {
		PRINTM(INFO, "Thread does not exist\n");
		return -1;
	}
#ifdef LINUX_2_4
	/* Acquire kernel lock */
	lock_kernel();

	/* Initialise wait mechanism */
	init_completion(&thr->sync);

	/* Reset the state to Stopped */
	thr->state = WLAN_THREAD_STOPPED;

	/* Remove the thread from the process table */
	kill_proc(thr->pid, SIGKILL, 1);

	/* Wait for thread to complete de-activation */
	wait_for_completion(&thr->sync);

	/* Release the kernel lock */
	unlock_kernel();

	/* Tell kernel to cleanup the thread which is in zombie state */
	kill_proc(2, SIGCHLD, 1);
#else
	kthread_stop(thr->task);
#endif

	LEAVE();
	return 0;
}

#endif
