/** @file  wlan_wext.c 
  * @brief This file contains ioctl functions
  * 
  *  Copyright  Marvell International Ltd. and/or its affiliates, 2003-2006
      This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
/********************************************************
Change log:
	10/10/05: Add Doxygen format comments
	12/23/05: Modify FindBSSIDInList to search entire table for
	          duplicate BSSIDs when earlier matches are not compatible
	12/26/05: Remove errant memcpy in wlanidle_off; overwriting stack space
	01/05/06: Add kernel 2.6.x support	
	01/11/06: Change compile flag BULVERDE_SDIO to SD to support Monahans/Zylonite
	01/11/06: Conditionalize new scan/join functions.
	          Update statics/externs.  Move forward decl. from wlan_decl.h
********************************************************/


#include	"include.h"

#define WAIT_FOR_SCAN_RRESULT_MAX_TIME (10 * HZ)
/********************************************************
		Local Variables
********************************************************/

/********************************************************
		Global Variables
********************************************************/


/********************************************************
		Local Functions
********************************************************/
static int wlan_set_rate(struct net_device *dev, struct iw_request_info *info, 
                         struct iw_param *vwrq, char *extra);
static int wlan_get_rate(struct net_device *dev, struct iw_request_info *info, 
                         struct iw_param *vwrq, char *extra);

static int wlan_get_essid(struct net_device *dev, struct iw_request_info *info,
                          struct iw_point *dwrq, char *extra);

static int wlan_set_freq(struct net_device *dev, struct iw_request_info *info, 
                         struct iw_freq *fwrq, char *extra);
static int wlan_get_freq(struct net_device *dev, struct iw_request_info *info, 
                         struct iw_freq *fwrq, char *extra);

static int wlan_set_mode(struct net_device *dev, struct iw_request_info *info, 
                         u32 * uwrq, char *extra);
static int wlan_get_mode(struct net_device *dev, struct iw_request_info *info,
                         u32 * uwrq, char *extra);

static int wlan_set_encode(struct net_device *dev,struct iw_request_info *info,
                           struct iw_point *dwrq, char *extra);
static int wlan_get_encode(struct net_device *dev,struct iw_request_info *info,
                           struct iw_point *dwrq, u8 * extra);

static int wlan_set_txpow(struct net_device *dev, struct iw_request_info *info,
                          struct iw_param *vwrq, char *extra);
static int wlan_get_txpow(struct net_device *dev, struct iw_request_info *info,
                          struct iw_param *vwrq, char *extra);


/** 
 *  @brief Find a character in a string.
 *   
 *  @param s	   A pointer to string
 *  @param c	   Character to be located 
 *  @param dlen    the length of string
 *  @return 	   A pointer to the first occurrence of c in string, or NULL if c is not found.
 */
static void *wlan_memchr(void *s, int c, int n)
{
	const u8	*p = s;

	while(n-- != 0) {
		if((u8)c == *p++) {
			return (void *) (p - 1);
		}
	}
	return NULL;
}


#if WIRELESS_EXT > 14
/** 
 *  @brief Convert mw value to dbm value
 *   
 *  @param mw	   the value of mw
 *  @return 	   the value of dbm
 */
static int mw_to_dbm(int mw)
{
	if (mw < 2)
		return 0;
	else if (mw < 3)
		return 3;
	else if (mw < 4)
		return 5;
	else if (mw < 6)
		return 7;
	else if (mw < 7)
		return 8;
	else if (mw < 8)
		return 9;
	else if (mw < 10)
		return 10;
	else if (mw < 13)
		return 11;
	else if (mw < 16)
		return 12;
	else if (mw < 20)
		return 13;
	else if (mw < 25)
		return 14;
	else if (mw < 32)
		return 15;
	else if (mw < 40)
		return 16;
	else if (mw < 50)
		return 17;
	else if (mw < 63)
		return 18;
	else if (mw < 79)
		return 19;
	else if (mw < 100)
		return 20;
	else
		return 21;
}
#endif

/** 
 *  @brief Find the channel frequency power info with specific channel
 *   
 *  @param adapter 	A pointer to wlan_adapter structure
 *  @param band		it can be BAND_A, BAND_G or BAND_B
 *  @param channel      the channel for looking	
 *  @return 	   	A pointer to CHANNEL_FREQ_POWER structure or NULL if not find.
 */
CHANNEL_FREQ_POWER *find_cfp_by_band_and_channel(wlan_adapter *adapter, 
                                                 u8 band, u16 channel)
{
	CHANNEL_FREQ_POWER	*cfp = NULL;
	REGION_CHANNEL		*rc;
	int	count = sizeof(adapter->region_channel) / 
				sizeof(adapter->region_channel[0]);
	int i, j;

	for (j = 0; !cfp && (j < count); j++) {
		rc = &adapter->region_channel[j];

		if( adapter->State11D.Enable11D == ENABLE_11D ) {
			rc = &adapter->universal_channel[j];
		}
		if(!rc->Valid || !rc->CFP)
			continue;
		if (rc->Band != band)
			continue;
		for (i = 0; i < rc->NrCFP; i++) {
			if (rc->CFP[i].Channel == channel) {
				cfp = &rc->CFP[i];
				break;
			}
		}
	}

	if (!cfp && channel)
		PRINTM(INFO, "find_cfp_by_band_and_channel(): cannot find "
			"cfp by band %d & channel %d\n",band,channel);

	return cfp;
}
			
/** 
 *  @brief Find the channel frequency power info with specific frequency
 *   
 *  @param adapter 	A pointer to wlan_adapter structure
 *  @param band		it can be BAND_A, BAND_G or BAND_B
 *  @param freq	        the frequency for looking	
 *  @return 	   	A pointer to CHANNEL_FREQ_POWER structure or NULL if not find.
 */
static CHANNEL_FREQ_POWER *find_cfp_by_band_and_freq(wlan_adapter *adapter, 
							u8 band, u32 freq)
{
	CHANNEL_FREQ_POWER *cfp = NULL;
	REGION_CHANNEL *rc;
	int count = sizeof(adapter->region_channel) / 
				sizeof(adapter->region_channel[0]);
	int i, j;

	for (j = 0; !cfp && (j < count); j++) {
		rc = &adapter->region_channel[j];

		if( adapter->State11D.Enable11D == ENABLE_11D ) {
			rc = &adapter->universal_channel[j];
		}

		if(!rc->Valid || !rc->CFP)
			continue;
		if (rc->Band != band)
			continue;
		for (i = 0; i < rc->NrCFP; i++) {
			if (rc->CFP[i].Freq == freq) {
				cfp = &rc->CFP[i];
				break;
			}
		}
	}

	if(!cfp && freq)
		PRINTM(INFO, "find_cfp_by_band_and_freql(): cannot find cfp by "
				"band %d & freq %d\n",band,freq);

	return cfp;
}
	

#ifdef MANF_CMD_SUPPORT
/** 
 *  @brief Manufacturing command ioctl function
 *   
 *  @param priv 	A pointer to wlan_private structure
 *  @param userdata	A pointer to user buf
 *  @return 	   	WLAN_STATUS_SUCCESS--success, WLAN_STATUS_FAILURE--fail
 */
static int wlan_mfg_command(wlan_private * priv, void *userdata)
{
	PkHeader	*pkHdr;
	int		len, ret;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	// creating the cmdbuf
	if (Adapter->mfg_cmd == NULL) {
		PRINTM(INFO, "Creating cmdbuf\n");
		if (!(Adapter->mfg_cmd = kmalloc(2000, GFP_KERNEL))) {
			PRINTM(INFO, "kmalloc failed!\n");
			return WLAN_STATUS_FAILURE;
		}
		Adapter->mfg_cmd_len = 2000;
	}
	
	// get PktHdr from userdata
	if (copy_from_user(Adapter->mfg_cmd, userdata, sizeof(PkHeader))) {
		PRINTM(INFO, "copy from user failed :PktHdr\n");
		return WLAN_STATUS_FAILURE;
	}

	// get the size
	pkHdr = (PkHeader *) Adapter->mfg_cmd;
	len = pkHdr->len;

	PRINTM(INFO, "cmdlen = %d\n", (u32) len);

	while (len >= Adapter->mfg_cmd_len) {
		kfree(Adapter->mfg_cmd);

		if (!(Adapter->mfg_cmd = kmalloc(len + 256, GFP_KERNEL))) {
			PRINTM(INFO, "kmalloc failed!\n");
			return WLAN_STATUS_FAILURE;
		}

		Adapter->mfg_cmd_len = len + 256;
	}

	// get the whole command from user
	if (copy_from_user(Adapter->mfg_cmd, userdata, len)) {
		PRINTM(INFO, "copy from user failed :PktHdr\n");
		return WLAN_STATUS_FAILURE;
	}

	ret = PrepareAndSendCommand(priv,
			HostCmd_CMD_MFG_COMMAND,
			0, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	if (ret) {
		LEAVE();
		return ret;
	}

	// copy it back to user
	if (Adapter->mfg_cmd_resp_len > 0) {
		if (copy_to_user(userdata, Adapter->mfg_cmd, len)) {
			PRINTM(INFO, "copy to user failed \n");
			return WLAN_STATUS_FAILURE;
		}
	}
	kfree(Adapter->mfg_cmd);
	Adapter->mfg_cmd = NULL;
	Adapter->mfg_cmd_len = 0;
	LEAVE();
	return ret;
}
#endif



/** 
 *  @brief Update Current Channel 
 *   
 *  @param priv 		A pointer to wlan_private structure
 *  @return 	   		WLAN_STATUS_SUCCESS--success, WLAN_STATUS_FAILURE--fail
 */
static int UpdateCurrentChannel(wlan_private *priv)
{
    int ret;
    
    /*
    ** the channel in f/w could be out of sync, get the current channel
    */
    ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_CHANNEL,
                                HostCmd_OPT_802_11_RF_CHANNEL_GET,
                                HostCmd_OPTION_WAITFORRSP, 0,
                                NULL);
    
    PRINTM(INFO, "Current Channel = %d\n", priv->adapter->Channel);

    return ret;
}

/** 
 *  @brief Set Current Channel 
 *   
 *  @param priv 		A pointer to wlan_private structure
 *  @param channel		The channel to be set. 
 *  @return 	   		WLAN_STATUS_SUCCESS--success, WLAN_STATUS_FAILURE--fail
 */
static int SetCurrentChannel(wlan_private *priv, int channel)
{
    PRINTM(INFO, "Set Channel = %d\n", channel);

    /* 
    **  Current channel is not set to AdhocChannel requested, set channel
    */
    return (PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_CHANNEL,
                                  HostCmd_OPT_802_11_RF_CHANNEL_SET,
                                  HostCmd_OPTION_WAITFORRSP, 0,
                                  &channel));
}

/** 
 *  @brief Change Adhoc Channel
 *   
 *  @param priv 		A pointer to wlan_private structure
 *  @param channel		The channel to be set. 
 *  @return 	   		WLAN_STATUS_SUCCESS--success, WLAN_STATUS_FAILURE--fail
 */
static int ChangeAdhocChannel(wlan_private *priv, int channel)
{
    	int             ret      = WLAN_STATUS_SUCCESS;
	wlan_adapter    *Adapter = priv->adapter;

    	Adapter->AdhocChannel = channel;

    	UpdateCurrentChannel(priv);

    	if (Adapter->Channel == Adapter->AdhocChannel){
        	/* AdhocChannel is set to the current Channel already */
        	LEAVE();
        	return WLAN_STATUS_SUCCESS;
    	}

    	PRINTM(INFO, "Updating Channel from %d to %d\n", 
	Adapter->Channel, Adapter->AdhocChannel);

    	SetCurrentChannel(priv, Adapter->AdhocChannel);

    	UpdateCurrentChannel(priv);

    	if (Adapter->Channel != Adapter->AdhocChannel){
        	PRINTM(INFO, "Failed to updated Channel to %d, channel = %d\n", 
                Adapter->AdhocChannel, Adapter->Channel);
        	LEAVE();
        	return WLAN_STATUS_FAILURE;
    	}

	if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		int			i;
		WLAN_802_11_SSID	curAdhocSsid;
		
		PRINTM(INFO, "Channel Changed while in an IBSS\n");

		/* Copy the current ssid */
		memcpy(&curAdhocSsid, &Adapter->CurBssParams.ssid, 
				sizeof(WLAN_802_11_SSID));
		
		/* Exit Adhoc mode */
		PRINTM(INFO, "In ChangeAdhocChannel(): Sending Adhoc Stop\n");
		ret = StopAdhocNetwork(priv);

		if (ret) {
			LEAVE();
			return ret;
		}

        /* Scan for the network, do not save previous results.  Stale
         *   scan data will cause us to join a non-existant adhoc network
         */
        SendSpecificSSIDScan(priv, &curAdhocSsid, FALSE);

		// find out the BSSID that matches the current SSID 
		i = FindSSIDInList(Adapter, &curAdhocSsid, NULL, 
				Wlan802_11IBSS);

		if (i >= 0) {
			PRINTM(INFO, "SSID found at %d in List," 
					"so join\n", i);
			JoinAdhocNetwork(priv, &Adapter->ScanTable[i]);
		} else {
			// else send START command
			PRINTM(INFO, "SSID not found in list, "
					"so creating adhoc with ssid = %s\n",
					curAdhocSsid.Ssid);
			StartAdhocNetwork(priv, &curAdhocSsid);
		}	// end of else (START command)
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}



/** 
 *  @brief Set WPA key
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_encode_wpa(struct net_device *dev, 
			struct iw_request_info *info,
				struct iw_point *dwrq, char *extra)
{
	int			ret = WLAN_STATUS_SUCCESS;
	wlan_private		*priv = dev->priv;
	wlan_adapter		*Adapter = priv->adapter;
	PWLAN_802_11_KEY	pKey;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "():%s IOCTLS called when station is"
					" in DeepSleep\n",__FUNCTION__);
		return -EBUSY;
	}

	pKey = (PWLAN_802_11_KEY) extra;

	HEXDUMP("Key buffer: ", extra, dwrq->length);

	HEXDUMP("KeyMaterial: ", (u8 *) pKey->KeyMaterial,
			pKey->KeyLength);

	// current driver only supports key length of up to 32 bytes
	if (pKey->KeyLength > MRVL_MAX_WPA_KEY_LENGTH) {
		PRINTM(INFO, " Error in key length \n");
		return WLAN_STATUS_FAILURE;
	}

	ret = PrepareAndSendCommand(priv,
		HostCmd_CMD_802_11_KEY_MATERIAL,
		HostCmd_ACT_SET,
		HostCmd_OPTION_WAITFORRSP,
		KEY_INFO_ENABLED, pKey);
	if (ret) {
		LEAVE();
		return ret;
	}

	LEAVE();
	return ret;
}

/*
 *  iwconfig ethX key on:	WEPEnabled;
 *  iwconfig ethX key off:	WEPDisabled;
 *  iwconfig ethX key [x]:	CurrentWepKeyIndex = x; WEPEnabled;
 *  iwconfig ethX key [x] kstr:	WepKey[x] = kstr;
 *  iwconfig ethX key kstr:	WepKey[CurrentWepKeyIndex] = kstr;
 *
 *  all:			Send command SET_WEP;
 				SetMacPacketFilter;
 */

/** 
 *  @brief Set WEP key
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_encode_nonwpa(struct net_device *dev,
			       struct iw_request_info *info,
				       struct iw_point *dwrq, char *extra)
{
	int			ret = WLAN_STATUS_SUCCESS;
	wlan_private		*priv = dev->priv;
	wlan_adapter		*Adapter = priv->adapter;
	MRVL_WEP_KEY		*pWep;
	WLAN_802_11_SSID	ssid;
	int			index, PrevAuthMode;

	ENTER();

	pWep = &Adapter->WepKey[Adapter->CurrentWepKeyIndex];
	PrevAuthMode = Adapter->SecInfo.AuthenticationMode;

	index = (dwrq->flags & IW_ENCODE_INDEX) - 1;

	if(index >= 4) {
		PRINTM(INFO, "Key index #%d out of range.\n", index + 1);
		return -EINVAL;
	}

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "():%s IOCTLS called when station is"
					" in DeepSleep\n",__FUNCTION__);
		return -EBUSY;
	}

	PRINTM(INFO, "Flags=0x%x, Length=%d Index=%d CurrentWepKeyIndex=%d\n", 
			dwrq->flags, dwrq->length, index, Adapter->CurrentWepKeyIndex);

	if (dwrq->length > 0) {
		/* iwconfig ethX key [n] xxxxxxxxxxx 
		 * Key has been provided by the user 
		 */

		/*
		 * Check the size of the key 
		 */

		if (dwrq->length > MAX_KEY_SIZE) {
			return -EINVAL;
		}

		/*
		 * Check the index (none -> use current) 
		 */

		if (index < 0 || index > 3)		//invalid index or no index
			index = Adapter->CurrentWepKeyIndex;
		else					//index is given & valid
			pWep = &Adapter->WepKey[index];

		/*
		 * Check if the key is not marked as invalid 
		 */
		if (!(dwrq->flags & IW_ENCODE_NOKEY)) {
			/* Cleanup */
			memset(pWep, 0, sizeof(MRVL_WEP_KEY));

			/* Copy the key in the driver */
			memcpy(pWep->KeyMaterial, extra, dwrq->length);

			/* Set the length */
			if (dwrq->length > MIN_KEY_SIZE) {
				pWep->KeyLength = MAX_KEY_SIZE;
			} else {
				if (dwrq->length > 0) {
					pWep->KeyLength = MIN_KEY_SIZE;
				} else {
					/* Disable the key */
					pWep->KeyLength = 0;
				}
			}
			pWep->KeyIndex = index;

			if(Adapter->SecInfo.WEPStatus != Wlan802_11WEPEnabled) {
				/*
				 * The status is set as Key Absent 
				 * so as to make sure we display the 
				 * keys when iwlist ethX key is 
				 * used - MPS 
				 */

				Adapter->SecInfo.WEPStatus = 
					Wlan802_11WEPKeyAbsent;
			}

			PRINTM(INFO, "KeyIndex=%u KeyLength=%u\n", 
					pWep->KeyIndex,
					pWep->KeyLength);
			HEXDUMP("WepKey", (u8 *) pWep->KeyMaterial, 
					pWep->KeyLength);
		}
	} else {
		/*
		 * No key provided so it is either enable key, 
		 * on or off */
		if (dwrq->flags & IW_ENCODE_DISABLED) {
			PRINTM(INFO, "******** iwconfig ethX key off **********\n");

			Adapter->SecInfo.WEPStatus = Wlan802_11WEPDisabled;
			if (Adapter->SecInfo.AuthenticationMode == Wlan802_11AuthModeShared)
				Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeOpen;
		} else {
			/* iwconfig ethX key [n]
			 * iwconfig ethX key on 
			 * Do we want to just set the transmit key index ? 
			 */

			if (index < 0 || index > 3) {
				PRINTM(INFO, "******** iwconfig ethX key on **********\n");
				index = Adapter->CurrentWepKeyIndex;
			}
			else {
				PRINTM(INFO, "******** iwconfig ethX key [x=%d] **********\n",index);
				Adapter->CurrentWepKeyIndex = index;
			}

			/* Copy the required key as the current key */
			pWep = &Adapter->WepKey[index];

			if (!pWep->KeyLength) {
				PRINTM(INFO, "Key not set,so cannot enable it\n");
				return -EPERM;
			}
	    
			Adapter->SecInfo.WEPStatus = Wlan802_11WEPEnabled;

			HEXDUMP("KeyMaterial", (u8 *) pWep->KeyMaterial,
					pWep->KeyLength);
		}
	}

	if(pWep->KeyLength) {
		ret = PrepareAndSendCommand(priv,
			HostCmd_CMD_802_11_SET_WEP,
			0, HostCmd_OPTION_WAITFORRSP,
			OID_802_11_ADD_WEP, NULL);

		if (ret) {
			LEAVE();
			return ret;
		}
	}

	if (Adapter->SecInfo.WEPStatus == Wlan802_11WEPEnabled) {
		Adapter->CurrentPacketFilter |= HostCmd_ACT_MAC_WEP_ENABLE;
	} else {
		Adapter->CurrentPacketFilter &= ~HostCmd_ACT_MAC_WEP_ENABLE;
	}

	SetMacPacketFilter(priv);

	if (dwrq->flags & IW_ENCODE_RESTRICTED) {
		/* iwconfig ethX restricted key [1] */
		Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeShared;
		PRINTM(INFO, "Authentication mode restricted!\n");
	} else if (dwrq->flags & IW_ENCODE_OPEN) {
		/* iwconfig ethX key [2] open */
		Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeOpen;
		PRINTM(INFO, "Authentication mode open!\n");
	}

	/*
	 * If authentication mode changed - de-authenticate, set authentication
	 * method and re-associate if we were previously associated.
	 */
	if (Adapter->SecInfo.AuthenticationMode != PrevAuthMode) {
		if (Adapter->MediaConnectStatus == WlanMediaStateConnected &&
				Adapter->InfrastructureMode == 
				Wlan802_11Infrastructure) {

			/* keep a copy of the ssid associated with */
			memcpy(&ssid, &Adapter->CurBssParams.ssid, sizeof(ssid));

			/*
			 * De-authenticate from AP 
			 */
				
			ret = SendDeauthentication(priv);

			if (ret) {
				LEAVE();
				return ret;
			}

		} else {
			/* reset ssid */
			memset(&ssid, 0, sizeof(ssid));
		}
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Set RX Antenna
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param Mode			RF antenna mode
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int SetRxAntenna(wlan_private *priv, int Mode)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_adapter	*Adapter = priv->adapter;

	if (Mode != RF_ANTENNA_1 && Mode != RF_ANTENNA_2
					&& Mode != RF_ANTENNA_AUTO) {
		return -EINVAL;
	}

	Adapter->RxAntennaMode = Mode;

	PRINTM(INFO, "SET RX Antenna Mode to 0x%04x\n", Adapter->RxAntennaMode);
			
	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_ANTENNA,
			HostCmd_ACT_SET_RX, HostCmd_OPTION_WAITFORRSP,
			0, &Adapter->RxAntennaMode);
	return ret;
}

/** 
 *  @brief Set TX Antenna
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param Mode			RF antenna mode
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int SetTxAntenna(wlan_private *priv, int Mode)
{
	int		ret = 0;
	wlan_adapter	*Adapter = priv->adapter;

	if ((Mode != RF_ANTENNA_1) && (Mode != RF_ANTENNA_2)
			&& (Mode != RF_ANTENNA_AUTO)) {
		return -EINVAL;
	}
		
	Adapter->TxAntennaMode = Mode; 

	PRINTM(INFO, "SET TX Antenna Mode to 0x%04x\n", Adapter->TxAntennaMode);

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_ANTENNA,
			HostCmd_ACT_SET_TX, HostCmd_OPTION_WAITFORRSP,
			0, &Adapter->TxAntennaMode);

	return ret;
}

/** 
 *  @brief Get RX Antenna
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param buf			A pointer to recieve antenna mode
 *  @return 	   		length of buf 
 */
static int GetRxAntenna(wlan_private *priv, char *buf)
{
	int		ret = 0;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	// clear it, so we will know if the value 
	// returned below is correct or not.
	Adapter->RxAntennaMode = 0;	

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_ANTENNA,
			HostCmd_ACT_GET_RX, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	if (ret) {
		LEAVE();
		return ret;
	}

	PRINTM(INFO, "Get Rx Antenna Mode:0x%04x\n", Adapter->RxAntennaMode);

	LEAVE();
			
	return sprintf(buf, "0x%04x", Adapter->RxAntennaMode) + 1;
}


/** 
 *  @brief Get TX Antenna
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param buf			A pointer to recieve antenna mode
 *  @return 	   		length of buf 
 */
static int GetTxAntenna(wlan_private *priv, char *buf)
{
	int		ret = 0;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();
	
	// clear it, so we will know if the value 
	// returned below is correct or not.
	Adapter->TxAntennaMode = 0;	

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_RF_ANTENNA,
			HostCmd_ACT_GET_TX, HostCmd_OPTION_WAITFORRSP,
			0, NULL);		

	if (ret) {
		LEAVE();
		return ret;
	}

	PRINTM(INFO, "Get Tx Antenna Mode:0x%04x\n", Adapter->TxAntennaMode);

	LEAVE();

	return sprintf(buf, "0x%04x", Adapter->TxAntennaMode) + 1;
}


/** 
 *  @brief Set Radio On/OFF
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @option 			Radio Option
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail 
 */
static int wlan_radio_ioctl(wlan_private *priv, u8 option)
{
	int		ret = 0;
	wlan_adapter	*Adapter	= priv->adapter;

	ENTER();

	if (Adapter->RadioOn != option) {
		PRINTM(INFO, "Switching %s the Radio\n", option ? "On" : "Off");
		Adapter->RadioOn = option;
		
		ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_RADIO_CONTROL,
				HostCmd_ACT_GEN_SET, HostCmd_OPTION_WAITFORRSP,
			      	0, NULL);
	}

	LEAVE();
	return ret;
}

#ifdef REASSOCIATION
/** 
 *  @brief Set Auto Reassociation On
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail 
 */
static int reassociation_on(wlan_private *priv)
{
	wlan_adapter	*Adapter	= priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "():%s IOCTLS called when station is"
					" in DeepSleep\n",__FUNCTION__);
		return -EBUSY;
	}

	Adapter->Reassoc_on = TRUE;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Auto Reassociation Off
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail 
 */
static int reassociation_off(wlan_private *priv)
{
	wlan_adapter		*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "():%s IOCTLS called when station is"
					" in DeepSleep\n",__FUNCTION__);
		return -EBUSY;
	}
	
	if (Adapter->TimerIsSet == TRUE) {
		CancelTimer(&Adapter->MrvDrvTimer);
		Adapter->TimerIsSet = FALSE;
	}	

	Adapter->Reassoc_on = FALSE;
	
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}
#endif /* REASSOCIATION */


/** 
 *  @brief Set Region
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param region_code		region code
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail 
 */
static int wlan_set_region(wlan_private *priv, u16 region_code)
{
	int i;

	ENTER();
	
	for (i = 0; i < MRVDRV_MAX_REGION_CODE; i++) {
		// use the region code to search for the index
		if (region_code == RegionCodeToIndex[i]) {
			priv->adapter->RegionTableIndex = (u16) i;
			priv->adapter->RegionCode = region_code;
			break;
		}
	}

	// if it's unidentified region code
	if (i >= MRVDRV_MAX_REGION_CODE) {
		PRINTM(INFO, "Region Code not identified\n");
		LEAVE();
		return WLAN_STATUS_FAILURE;
	}

	if (wlan_set_regiontable(priv, priv->adapter->RegionCode, 0)) {
		LEAVE();
		return -EINVAL;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Copy Rates
 *   
 *  @param dest                 A pointer to Dest Buf
 *  @param src		        A pointer to Src Buf
 *  @param len                  The len of Src Buf
 *  @return 	   	        Number of Rates copyed 
 */
static inline int CopyRates(u8 *dest, int pos, u8 *src, int len)
{
	int	i;

	for (i = 0; i < len && src[i]; i++, pos++) {
		if (pos >= sizeof(WLAN_802_11_RATES))
			break;
		dest[pos] = src[i];
	}

	return pos;
}

/** 
 *  @brief Get active data rates
 *   
 *  @param Adapter              A pointer to wlan_adapter structure
 *  @param rate		        The buf to return the active rates
 *  @return 	   	        The number of Rates
 */
static int get_active_data_rates(wlan_adapter *Adapter, WLAN_802_11_RATES rates)
{
	int	k = 0;

	ENTER();
	
	if (Adapter->MediaConnectStatus != WlanMediaStateConnected) {
		if (Adapter->InfrastructureMode == Wlan802_11Infrastructure) {	
			//Infra. mode
			PRINTM(INFO, "Infra\n");
			k = CopyRates(rates, k, SupportedRates, 
						sizeof(SupportedRates));
		} else {
			//ad-hoc mode
			PRINTM(INFO, "Adhoc G\n");
			k = CopyRates(rates, k, AdhocRates_G, 
							sizeof(AdhocRates_G));
		}
	} else {
		k = CopyRates(rates, 0, Adapter->CurBssParams.DataRates, 
					Adapter->CurBssParams.NumOfRates);
	}

	LEAVE();

	return k;
}

/** 
 *  @brief Get/Set Firmware wakeup method
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_txcontrol(wlan_private *priv, struct iwreq *wrq)
{
	wlan_adapter	*Adapter = priv->adapter;
	int data;	
	ENTER();

	if((int) wrq->u.data.length == 0) {
	    if(copy_to_user(wrq->u.data.pointer, &Adapter->PktTxCtrl, sizeof(u32))) {
		PRINTM(MSG, "copy_to_user failed!\n");
		return -EFAULT;
	    }	    
	} else{
	     if((int) wrq->u.data.length > 1) {
	  	PRINTM(MSG, "ioctl too many args!\n");
	   	return -EFAULT;
	     }
	     copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
	     (u32)Adapter->PktTxCtrl = data;
	}

	wrq->u.data.length = 1;
			
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Enable/Disable atim uapsd null package generation
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_null_pkg_gen( wlan_private *priv, struct iwreq *wrq )
{
	int data;
	wlan_adapter   *Adapter = priv->adapter;
	
	ENTER();

	if ( wrq->u.data.length == 0 ) {
		return -EINVAL;
	}
	copy_from_user(&data,wrq->u.data.pointer,sizeof(int) );

	PRINTM(INFO, "Enable UAPSD NULL PKG: %s\n", 
			(data==CMD_ENABLED)?"Enable":"Disable");
	
	switch ( data ) {
		case CMD_ENABLED:
			Adapter->gen_null_pkg = TRUE;
			break;
		case CMD_DISABLED:
			Adapter->gen_null_pkg = FALSE;
			break;
		default:			
			break;
	}

	data = (Adapter->gen_null_pkg ==TRUE)?CMD_ENABLED:CMD_DISABLED;
	copy_to_user(wrq->u.data.pointer, &data, sizeof(int) );
	wrq->u.data.length = 1;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get/Set NULL Package generation interval
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_null_pkt_interval(wlan_private *priv, struct iwreq *wrq)
{
	wlan_adapter	*Adapter = priv->adapter;
	int data;	
	ENTER();

	if((int) wrq->u.data.length == 0) {
	    data = Adapter->NullPktInterval;
	    
	    if(copy_to_user(wrq->u.data.pointer, &data, sizeof(int))) {
		PRINTM(MSG, "copy_to_user failed!\n");
		return -EFAULT;
	    }
	} else{
	     if((int) wrq->u.data.length > 1) {
	  	PRINTM(MSG, "ioctl too many args!\n");
	   	return -EFAULT;
	     }
	     copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
	     Adapter->NullPktInterval = data;
	}

	wrq->u.data.length = 1;
			
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


#ifdef	linux
#include	"wlan_version.h"

#define GETLOG_BUFSIZE  300

#define MAX_SCAN_CELL_SIZE      (IW_EV_ADDR_LEN + \
				MRVDRV_MAX_SSID_LENGTH + \
				IW_EV_UINT_LEN + IW_EV_FREQ_LEN + \
				IW_EV_QUAL_LEN + MRVDRV_MAX_SSID_LENGTH + \
				IW_EV_PARAM_LEN + 40) /* 40 for WPAIE */



/** 
 *  @brief Commit handler: called after a bunch of SET operations
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_config_commit(struct net_device *dev, 
		struct iw_request_info *info, 
		char *cwrq, char *extra)
{
	ENTER();

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Get protocol name 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_name(struct net_device *dev, struct iw_request_info *info, 
		char *cwrq, char *extra)
{
	const char *cp;
	char comm[6] = {"COMM-"};
	char mrvl[6] = {"MRVL-"};
	int cnt;

	ENTER();

	strcpy(cwrq, mrvl);

	cp = strstr(driver_version, comm);
	if (cp == driver_version)	//skip leading "COMM-"
		cp = driver_version + strlen(comm);
	else
		cp = driver_version;

	cnt = strlen(mrvl);
	cwrq += cnt;
	while (cnt < 16 && (*cp != '-')) {
		*cwrq++ = toupper(*cp++);
		cnt++;
	}
	*cwrq = '\0';

	LEAVE();

	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Get frequency
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param fwrq 		A pointer to iw_freq structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_freq(struct net_device *dev, struct iw_request_info *info, 
		struct iw_freq *fwrq, char *extra)
{
	wlan_private		*priv = dev->priv;
	wlan_adapter		*Adapter = priv->adapter;
	CHANNEL_FREQ_POWER	*cfp;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCGIWFREQ: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	cfp = find_cfp_by_band_and_channel(Adapter, 0, 
						Adapter->CurBssParams.channel);
	
	if (!cfp) {
		if (Adapter->CurBssParams.channel)
			PRINTM(INFO, "Invalid channel=%d\n", 
					Adapter->CurBssParams.channel);
		return -EINVAL;
	} 
	
	fwrq->m = (long) cfp->Freq * 100000;
	fwrq->e = 1;

	PRINTM(INFO, "freq=%u\n",fwrq->m);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}



/** 
 *  @brief Get current BSSID
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param awrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_wap(struct net_device *dev, struct iw_request_info *info, 
		struct sockaddr *awrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		memcpy(awrq->sa_data, Adapter->CurBssParams.bssid, ETH_ALEN);
	} else {
		memset(awrq->sa_data, 0, ETH_ALEN);
	}
	awrq->sa_family = ARPHRD_ETHER;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Adapter Node Name
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_nick(struct net_device *dev, struct iw_request_info *info, 
		struct iw_point *dwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	/*
	 * Check the size of the string 
	 */
	
	if (dwrq->length > 16) {
		return -E2BIG;
	}

	memset(Adapter->nodeName, 0, sizeof(Adapter->nodeName));
	memcpy(Adapter->nodeName, extra, dwrq->length);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get Adapter Node Name
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_nick(struct net_device *dev, struct iw_request_info *info, 
		struct iw_point *dwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	/*
	 * Get the Nick Name saved 
	 */

	strncpy(extra, Adapter->nodeName, 16);

	extra[16] = '\0';

	/*
	 * If none, we may want to get the one that was set 
	 */

	/*
	 * Push it out ! 
	 */
	dwrq->length = strlen(extra) + 1;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set RTS threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_rts(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	int		rthr = vwrq->value;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWRTS: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	if (vwrq->disabled) {
		Adapter->RTSThsd = rthr = MRVDRV_RTS_MAX_VALUE;
	} else {
		if (rthr < MRVDRV_RTS_MIN_VALUE || rthr > MRVDRV_RTS_MAX_VALUE)
			return -EINVAL;
		Adapter->RTSThsd = rthr;
	}

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SNMP_MIB,
			HostCmd_ACT_SET, HostCmd_OPTION_WAITFORRSP,
			OID_802_11_RTS_THRESHOLD, &rthr);

	LEAVE();
	return ret;
}


/** 
 *  @brief Get RTS threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_rts(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCGIWRTS: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	Adapter->RTSThsd = 0;
	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SNMP_MIB,
			HostCmd_ACT_GET, HostCmd_OPTION_WAITFORRSP,
			OID_802_11_RTS_THRESHOLD, NULL);
	if (ret) {
		LEAVE();
		return ret;
	}
	

	vwrq->value = Adapter->RTSThsd;
	vwrq->disabled = ((vwrq->value < MRVDRV_RTS_MIN_VALUE)
			 || (vwrq->value > MRVDRV_RTS_MAX_VALUE));
	vwrq->fixed = 1;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Fragment threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_frag(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	int		fthr = vwrq->value;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWFRAG: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	if (vwrq->disabled) {
		Adapter->FragThsd = fthr = MRVDRV_FRAG_MAX_VALUE;
	} else {
		if (fthr < MRVDRV_FRAG_MIN_VALUE || fthr > MRVDRV_FRAG_MAX_VALUE)
			return -EINVAL;
		Adapter->FragThsd = fthr;
	}

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SNMP_MIB,
			HostCmd_ACT_SET, HostCmd_OPTION_WAITFORRSP,
			OID_802_11_FRAGMENTATION_THRESHOLD,
			&fthr);
	LEAVE();
	return ret;
}

/** 
 *  @brief Get Fragment threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_frag(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCGIWFRAG: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}
	Adapter->FragThsd = 0;
	ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_SNMP_MIB, HostCmd_ACT_GET,
				HostCmd_OPTION_WAITFORRSP,
				OID_802_11_FRAGMENTATION_THRESHOLD,
				NULL);
	if (ret) {
		LEAVE();
		return ret;
	}

	vwrq->value = Adapter->FragThsd;
	vwrq->disabled = ((vwrq->value < MRVDRV_FRAG_MIN_VALUE)
			 || (vwrq->value > MRVDRV_FRAG_MAX_VALUE));
	vwrq->fixed = 1;

	LEAVE();
	return ret;
}

/** 
 *  @brief Get Wlan Mode
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_mode(struct net_device *dev,
	      struct iw_request_info *info, u32 * uwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*adapter = priv->adapter;

	ENTER();

	switch (adapter->InfrastructureMode) {
		case Wlan802_11IBSS:
			*uwrq = IW_MODE_ADHOC;
			break;

		case Wlan802_11Infrastructure:
			*uwrq = IW_MODE_INFRA;
			break;

		default:
		case Wlan802_11AutoUnknown:
			*uwrq = IW_MODE_AUTO;
			break;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Get Encryption key
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_encode(struct net_device *dev,
			struct iw_request_info *info,
				struct iw_point *dwrq, u8 * extra)
{

	wlan_private	*priv = dev->priv;
	wlan_adapter	*adapter = priv->adapter;
	int		index = (dwrq->flags & IW_ENCODE_INDEX);

	ENTER();

	PRINTM(INFO, "flags=0x%x index=%d length=%d CurrentWepKeyIndex=%d\n", 
			dwrq->flags, index, dwrq->length, 
			adapter->CurrentWepKeyIndex);

	dwrq->flags = 0;

	/*
	 * Check encryption mode 
	 */

	switch (adapter->SecInfo.AuthenticationMode) {
		case Wlan802_11AuthModeOpen:
			dwrq->flags = IW_ENCODE_OPEN;
			break;

		case Wlan802_11AuthModeShared:
		case Wlan802_11AuthModeNetworkEAP:
			dwrq->flags = IW_ENCODE_RESTRICTED;
			break;
		default:
			dwrq->flags = IW_ENCODE_DISABLED | IW_ENCODE_OPEN;
			break;
	}

	if ((adapter->SecInfo.WEPStatus == Wlan802_11WEPEnabled)
		|| (adapter->SecInfo.WEPStatus == Wlan802_11WEPKeyAbsent)
		|| adapter->SecInfo.WPAEnabled
		|| adapter->SecInfo.WPA2Enabled
		) {
		dwrq->flags &= ~IW_ENCODE_DISABLED;
	} else {
		dwrq->flags |= IW_ENCODE_DISABLED;
	}

	memset(extra, 0, 16);

	if (!index) {
		/* Handle current key request	*/
		if ((adapter->WepKey[adapter->CurrentWepKeyIndex].KeyLength) &&
				(adapter->SecInfo.WEPStatus == Wlan802_11WEPEnabled)) {
			index = adapter->WepKey[adapter->CurrentWepKeyIndex].KeyIndex;
			memcpy(extra, adapter->WepKey[index].KeyMaterial,
					adapter->WepKey[index].KeyLength);
			dwrq->length = adapter->WepKey[index].KeyLength;
			/* return current key */
			dwrq->flags |= (index + 1);
			/* return WEP enabled */
			dwrq->flags &= ~IW_ENCODE_DISABLED;
		} else if ((adapter->SecInfo.WPAEnabled)
		|| (adapter->SecInfo.WPA2Enabled)
		) {
			/* return WPA enabled */
			dwrq->flags &= ~IW_ENCODE_DISABLED;
		} else {
			dwrq->flags |= IW_ENCODE_DISABLED;
		}
	} else {
		/* Handle specific key requests */
		index--;
		if (adapter->WepKey[index].KeyLength) {
			memcpy(extra, adapter->WepKey[index].KeyMaterial,
					adapter->WepKey[index].KeyLength);
			dwrq->length = adapter->WepKey[index].KeyLength;
			/* return current key */
			dwrq->flags |= (index + 1);
			/* return WEP enabled */
			dwrq->flags &= ~IW_ENCODE_DISABLED;
		} else if ((adapter->SecInfo.WPAEnabled)
		|| (adapter->SecInfo.WPA2Enabled)
		) {
			/* return WPA enabled */
			dwrq->flags &= ~IW_ENCODE_DISABLED;
		} else {
			dwrq->flags |= IW_ENCODE_DISABLED;
		}
	}

	dwrq->flags |= IW_ENCODE_NOKEY;

	PRINTM(INFO, "Key:%02x:%02x:%02x:%02x:%02x:%02x KeyLen=%d\n",
			extra[0], extra[1], extra[2],
			extra[3], extra[4], extra[5], dwrq->length);

	if (adapter->EncryptionStatus == Wlan802_11Encryption2Enabled
			&& !dwrq->length) {
		dwrq->length = MAX_KEY_SIZE;
	}

	PRINTM(INFO, "Return flags=0x%x\n", dwrq->flags);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Get TX Power
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_txpow(struct net_device *dev,
			   struct iw_request_info *info,
			   struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCGIWTXPOW: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	ret = PrepareAndSendCommand(priv, 
			HostCmd_CMD_802_11_RF_TX_POWER,
			HostCmd_ACT_TX_POWER_OPT_GET,
			HostCmd_OPTION_WAITFORRSP,
		      	0, NULL);

	if (ret) {
		LEAVE();
		return ret;
	}

  	PRINTM(INFO, "TXPOWER GET %d dbm.\n", Adapter->TxPowerLevel);
	vwrq->value = Adapter->TxPowerLevel;
	vwrq->fixed = 1;
	if (Adapter->RadioOn) {
		vwrq->disabled = 0;
		vwrq->flags = IW_TXPOW_DBM;
	} else {
		vwrq->disabled = 1;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Set TX Retry Count
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_retry(struct net_device *dev, struct iw_request_info *info,
	       struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private   *priv = dev->priv;
	wlan_adapter   *adapter = priv->adapter;

	ENTER();

	if ((adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWRETRY: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	if (vwrq->flags == IW_RETRY_LIMIT) {
		/* The MAC has a 4-bit Total_Tx_Count register
		   Total_Tx_Count = 1 + Tx_Retry_Count */
#define TX_RETRY_MIN 0
#define TX_RETRY_MAX 14
		if (vwrq->value < TX_RETRY_MIN || vwrq->value > TX_RETRY_MAX)
			return -EINVAL;

		/* Adding 1 to convert retry count to try count */
		adapter->TxRetryCount = vwrq->value + 1;

		ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SNMP_MIB, 
				HostCmd_ACT_SET, HostCmd_OPTION_WAITFORRSP,
				OID_802_11_TX_RETRYCOUNT,
				NULL);

		if (ret) {
			LEAVE();
			return ret;
		}
	} else {
		return -EOPNOTSUPP;
	}

	LEAVE();
	return  WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get TX Retry Count
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_retry(struct net_device *dev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	wlan_private   *priv = dev->priv;
	wlan_adapter   *Adapter = priv->adapter;
	int		ret = WLAN_STATUS_SUCCESS;

	ENTER();
	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCGIWFRAG: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}
	Adapter->TxRetryCount = 0;
	ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_SNMP_MIB, HostCmd_ACT_GET,
				HostCmd_OPTION_WAITFORRSP,
				OID_802_11_TX_RETRYCOUNT,
				NULL);
	if (ret) {
		LEAVE();
		return ret;
	}
	vwrq->disabled = 0;
	if (!vwrq->flags) {
		vwrq->flags = IW_RETRY_LIMIT;
		/* Subtract 1 to convert try count to retry count */
		vwrq->value = Adapter->TxRetryCount - 1;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Sort Channels
 *   
 *  @param freq 		A pointer to iw_freq structure
 *  @param num		        number of Channels
 *  @return 	   		NA
 */
static inline void sort_channels(struct iw_freq *freq, int num)
{
	int i, j;
	struct iw_freq temp;

	for (i = 0; i < num; i++)
		for (j = i + 1; j < num; j++)
			if (freq[i].i > freq[j].i) {
				temp.i = freq[i].i;
				temp.m = freq[i].m;
				
				freq[i].i = freq[j].i;
				freq[i].m = freq[j].m;
				
				freq[j].i = temp.i;	
				freq[j].m = temp.m;	
			}
}

/* data rate listing
	MULTI_BANDS:
		abg		a	b	b/g
   Infra 	G(12)		A(8)	B(4)	G(12)
   Adhoc 	A+B(12)		A(8)	B(4)	B(4)

	non-MULTI_BANDS:
		   		 	b	b/g
   Infra 	     		    	B(4)	G(12)
   Adhoc 	      		    	B(4)	B(4)
 */
/** 
 *  @brief Get Range Info
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_range(struct net_device *dev, struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{
	int			i, j;
	wlan_private		*priv = dev->priv;
	wlan_adapter		*Adapter = priv->adapter;
	struct iw_range		*range = (struct iw_range *) extra;
	CHANNEL_FREQ_POWER	*cfp;
	WLAN_802_11_RATES	rates;

	BOOLEAN		flag = FALSE;

	ENTER();

	dwrq->length = sizeof(struct iw_range);
	memset(range, 0, sizeof(struct iw_range));

	range->min_nwid = 0;
	range->max_nwid = 0;

	memset(rates, 0, sizeof(rates));
	range->num_bitrates = get_active_data_rates(Adapter, rates);

	for (i=0; i < MIN(range->num_bitrates,IW_MAX_BITRATES) && rates[i]; i++) {
		range->bitrate[i] = (rates[i] & 0x7f) * 500000;
	}
	range->num_bitrates = i;
	PRINTM(INFO, "IW_MAX_BITRATES=%d num_bitrates=%d\n",IW_MAX_BITRATES,range->num_bitrates);

	range->num_frequency = 0;
	if( wlan_get_state_11d( priv) == ENABLE_11D &&
			Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		u8 chan_no;
		u8 band;
		
		parsed_region_chan_11d_t *parsed_region_chan = &Adapter->parsed_region_chan;

		if( parsed_region_chan == NULL ) {
			PRINTM(INFO, "11D:parsed_region_chan is NULL\n");
			LEAVE();
			return 0;
		}
		band = parsed_region_chan->band;
		PRINTM(INFO, "band=%d NoOfChan=%d\n", band, parsed_region_chan->NoOfChan);

		for (i = 0; (range->num_frequency < IW_MAX_FREQUENCIES) 
				&& (i < parsed_region_chan->NoOfChan); i++) {
			chan_no = parsed_region_chan->chanPwr[i].chan;
			PRINTM(INFO, "chan_no=%d\n", chan_no );
			range->freq[range->num_frequency].i = (long) chan_no;
			range->freq[range->num_frequency].m = 
					(long) chan_2_freq(chan_no, band) * 100000;
			range->freq[range->num_frequency].e = 1;
			range->num_frequency++;
		}
		flag = TRUE;
	}
	if ( !flag )
	{
		for (j = 0; (range->num_frequency < IW_MAX_FREQUENCIES) 
			&& (j < sizeof(Adapter->region_channel) 
				/ sizeof(Adapter->region_channel[0]));
				j++) {
			cfp = Adapter->region_channel[j].CFP;
			for (i = 0; (range->num_frequency < IW_MAX_FREQUENCIES) 
				&& Adapter->region_channel[j].Valid 
				&& cfp 
				&& (i < Adapter->region_channel[j].NrCFP);
					i++) {
				range->freq[range->num_frequency].i = (long) cfp->Channel;
				range->freq[range->num_frequency].m = (long) cfp->Freq * 100000;
				range->freq[range->num_frequency].e = 1;
				cfp++;
				range->num_frequency++;
			}
		}
	}

	PRINTM(INFO, "IW_MAX_FREQUENCIES=%d num_frequency=%d\n",IW_MAX_FREQUENCIES,
							range->num_frequency);

	range->num_channels = range->num_frequency;
	
	sort_channels(&range->freq[0], range->num_frequency);
	
	/*
	 * Set an indication of the max TCP throughput in bit/s that we can
	 * expect using this interface 
	 */
	if (i > 2)
		range->throughput = 5000 * 1000;
	else
		range->throughput = 1500 * 1000;

	range->min_rts = MRVDRV_RTS_MIN_VALUE;
	range->max_rts = MRVDRV_RTS_MAX_VALUE;
	range->min_frag = MRVDRV_FRAG_MIN_VALUE;
	range->max_frag = MRVDRV_FRAG_MAX_VALUE;

	range->encoding_size[0] = 5;
	range->encoding_size[1] = 13;
	range->num_encoding_sizes = 2;
	range->max_encoding_tokens = 4;

	range->min_pmp = 1000000;
	range->max_pmp = 120000000;	
	range->min_pmt = 1000;		
	range->max_pmt = 1000000;	
	range->pmp_flags = IW_POWER_PERIOD;
	range->pmt_flags = IW_POWER_TIMEOUT;
	range->pm_capa = IW_POWER_PERIOD | IW_POWER_TIMEOUT | IW_POWER_ALL_R;

	/*
	 * Minimum version we recommend 
	 */
	range->we_version_source = 15;

	/*
	 * Version we are compiled with 
	 */
	range->we_version_compiled = WIRELESS_EXT;

	range->retry_capa = IW_RETRY_LIMIT;
	range->retry_flags = IW_RETRY_LIMIT | IW_RETRY_MAX;

	range->min_retry = TX_RETRY_MIN;
	range->max_retry = TX_RETRY_MAX;

	/*
	 * Set the qual, level and noise range values 
	 */
	/*
	 * need to put the right values here 
	 */
	range->max_qual.qual = 10;
	range->max_qual.level = 0;
	range->max_qual.noise = 0;
	range->sensitivity = 0;

	/*
	 * Setup the supported power level ranges 
	 */
	memset(range->txpower, 0, sizeof(range->txpower));
	range->txpower[0] = 5;
	range->txpower[1] = 7;
	range->txpower[2] = 9;
	range->txpower[3] = 11;
	range->txpower[4] = 13;
	range->txpower[5] = 15;
	range->txpower[6] = 17;
	range->txpower[7] = 19;

	range->num_txpower = 8;
	range->txpower_capa = IW_TXPOW_DBM;
	range->txpower_capa |= IW_TXPOW_RANGE;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief  Set power management 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_power(struct net_device *dev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWPOWER: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	/* PS is currently supported only in Infrastructure Mode 
	 * Remove this check if it is to be supported in IBSS mode also 
	 */
	if (Adapter->InfrastructureMode != Wlan802_11Infrastructure)
		return -EINVAL;

	if (vwrq->disabled) {
		Adapter->PSMode = Wlan802_11PowerModeCAM;
		if (Adapter->PSState != PS_STATE_FULL_POWER) {
			PSWakeup(priv, HostCmd_OPTION_WAITFORRSP);
		}

		return 0;
	}

	if ((vwrq->flags & IW_POWER_TYPE) == IW_POWER_TIMEOUT) {
		PRINTM(INFO, "Setting power timeout command is not supported\n");
		return -EINVAL;
	} else if ((vwrq->flags & IW_POWER_TYPE) == IW_POWER_PERIOD) {
		PRINTM(INFO, "Setting power period command is not supported\n");
		return -EINVAL;
	}

	if (Adapter->PSMode != Wlan802_11PowerModeCAM) {
		return WLAN_STATUS_SUCCESS;
	}
     
	Adapter->PSMode = Wlan802_11PowerModeMAX_PSP;
     
	if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		PSSleep(priv, HostCmd_OPTION_WAITFORRSP);
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief  Get power management 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_power(struct net_device *dev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	int		mode;

	ENTER();

	mode = Adapter->PSMode;

	if ((vwrq->disabled = (mode == Wlan802_11PowerModeCAM))
			|| Adapter->MediaConnectStatus == 
					WlanMediaStateDisconnected) {
		LEAVE();
		return WLAN_STATUS_SUCCESS;
	}

	vwrq->value = 0;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief  Set sensitivity threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_sens(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	ENTER();
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief  Get sensitivity threshold
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_FAILURE
 */
static int wlan_get_sens(struct net_device *dev,
	      struct iw_request_info *info, struct iw_param *vwrq,
	      char *extra)
{
	ENTER();
	LEAVE();
	return WLAN_STATUS_FAILURE;
}




/*
 * iwconfig settable callbacks 
 */
static const iw_handler wlan_handler[] = {
    (iw_handler) wlan_config_commit,        /* SIOCSIWCOMMIT */
    (iw_handler) wlan_get_name,             /* SIOCGIWNAME */
    (iw_handler) NULL,                      /* SIOCSIWNWID */
    (iw_handler) NULL,                      /* SIOCGIWNWID */
    (iw_handler) wlan_set_freq,             /* SIOCSIWFREQ */
    (iw_handler) wlan_get_freq,             /* SIOCGIWFREQ */
    (iw_handler) wlan_set_mode,             /* SIOCSIWMODE */
    (iw_handler) wlan_get_mode,             /* SIOCGIWMODE */
    (iw_handler) wlan_set_sens,             /* SIOCSIWSENS */
    (iw_handler) wlan_get_sens,             /* SIOCGIWSENS */
    (iw_handler) NULL,                      /* SIOCSIWRANGE */
    (iw_handler) wlan_get_range,            /* SIOCGIWRANGE */
    (iw_handler) NULL,                      /* SIOCSIWPRIV */
    (iw_handler) NULL,                      /* SIOCGIWPRIV */
    (iw_handler) NULL,                      /* SIOCSIWSTATS */
    (iw_handler) NULL,                      /* SIOCGIWSTATS */
#if WIRELESS_EXT > 15
    iw_handler_set_spy,                     /* SIOCSIWSPY */
    iw_handler_get_spy,                     /* SIOCGIWSPY */
    iw_handler_set_thrspy,                  /* SIOCSIWTHRSPY */
    iw_handler_get_thrspy,                  /* SIOCGIWTHRSPY */
#else  /* WIRELESS_EXT > 15 */
#ifdef WIRELESS_SPY
    (iw_handler) wlan_set_spy,              /* SIOCSIWSPY */
    (iw_handler) wlan_get_spy,              /* SIOCGIWSPY */
#else  /* WIRELESS_SPY */
    (iw_handler) NULL,                      /* SIOCSIWSPY */
    (iw_handler) NULL,                      /* SIOCGIWSPY */
#endif /* WIRELESS_SPY */
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) NULL,                      /* -- hole -- */
#endif /* WIRELESS_EXT > 15 */
    (iw_handler) wlan_set_wap,              /* SIOCSIWAP */
    (iw_handler) wlan_get_wap,              /* SIOCGIWAP */
    (iw_handler) NULL,                      /* -- hole -- */
    //(iw_handler) wlan_get_aplist,         /* SIOCGIWAPLIST */
    NULL,                                   /* SIOCGIWAPLIST */
#if WIRELESS_EXT > 13
    (iw_handler) wlan_set_scan,             /* SIOCSIWSCAN */
    (iw_handler) wlan_get_scan,             /* SIOCGIWSCAN */
#else  /* WIRELESS_EXT > 13 */
    (iw_handler) NULL,                      /* SIOCSIWSCAN */
    (iw_handler) NULL,                      /* SIOCGIWSCAN */
#endif /* WIRELESS_EXT > 13 */
    (iw_handler) wlan_set_essid,            /* SIOCSIWESSID */
    (iw_handler) wlan_get_essid,            /* SIOCGIWESSID */
    (iw_handler) wlan_set_nick,             /* SIOCSIWNICKN */
    (iw_handler) wlan_get_nick,             /* SIOCGIWNICKN */
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) wlan_set_rate,             /* SIOCSIWRATE */
    (iw_handler) wlan_get_rate,             /* SIOCGIWRATE */
    (iw_handler) wlan_set_rts,              /* SIOCSIWRTS */
    (iw_handler) wlan_get_rts,              /* SIOCGIWRTS */
    (iw_handler) wlan_set_frag,             /* SIOCSIWFRAG */
    (iw_handler) wlan_get_frag,             /* SIOCGIWFRAG */
    (iw_handler) wlan_set_txpow,            /* SIOCSIWTXPOW */
    (iw_handler) wlan_get_txpow,            /* SIOCGIWTXPOW */
    (iw_handler) wlan_set_retry,            /* SIOCSIWRETRY */
    (iw_handler) wlan_get_retry,            /* SIOCGIWRETRY */
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) wlan_get_encode,           /* SIOCGIWENCODE */
    (iw_handler) wlan_set_power,            /* SIOCSIWPOWER */
    (iw_handler) wlan_get_power,            /* SIOCGIWPOWER */
#if (WIRELESS_EXT >= 18)                    
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) NULL,                      /* -- hole -- */
#if defined(IWGENIE_SUPPORT)                
    (iw_handler) wlan_set_gen_ie,           /* SIOCSIWGENIE */
    (iw_handler) wlan_get_gen_ie,           /* SIOCGIWGENIE */
#else /* IWGENIE_SUPPORT */                 
    (iw_handler) NULL,                      /* -- hole -- */
    (iw_handler) NULL,                      /* -- hole -- */
#endif /* IWGENIE_SUPPORT */
#endif /* WIRELESSS_EXT >= 18 */
};

/*
 * iwpriv settable callbacks 
 */

static const iw_handler wlan_private_handler[] = {
	NULL,			/* SIOCIWFIRSTPRIV */
};

static const struct iw_priv_args wlan_private_args[] = {
	/*
	 * { cmd, set_args, get_args, name } 
	 */
	{
	WLANEXTSCAN,
	IW_PRIV_TYPE_INT,
	IW_PRIV_TYPE_CHAR | 2,
	"extscan"
	},

	{
	WLANCISDUMP,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE | 512,
	"getcis"
	},

     	{
	WLANSCAN_TYPE,
	IW_PRIV_TYPE_CHAR | 8,
	IW_PRIV_TYPE_CHAR | 8,
	"scantype"
	},

	{ 
	WLAN_SETONEINT_GETONEINT,
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	""
	},
	{ 
	WLANNF, 
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	"getNF"
	},
	{ 
	WLANRSSI, 
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	"getRSSI"
	},

	{ 
	WLANWMM_ENABLE,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"wmm"
	},

	{ 
	WLANBGSCAN, 
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	"bgscan"
	},


	{
	WLANENABLE11D,		
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	"enable11d"
	},

	{ 
	WLANADHOCGRATE,
	IW_PRIV_TYPE_INT | 1, 
	IW_PRIV_TYPE_INT | 1, 
	"adhocgrate"
	},

	{
	WLAN_WMM_QOSINFO,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"wmm_qosinfo"
	},
	{
	WLAN_LISTENINTRVL,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"lolisteninter"
	},
	{
	WLAN_FW_WAKEUP_METHOD,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"fwwakeupmethod"
	},

	{
	WLAN_TXCONTROL,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"txcontrol"
	},

	{
	WLANNULLGEN,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"uapsdnullgen"
	},
	{
	WLAN_NULLPKTINTERVAL,
	IW_PRIV_TYPE_INT | 1,
	IW_PRIV_TYPE_INT | 1,
	"psnullinterval"
	},
	/* Using iwpriv sub-command feature */
	{ 
	WLAN_SETONEINT_GETNONE,  /* IOCTL: 24 */
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE, 
	""
	},

	{ 
	WLAN_SUBCMD_SETRXANTENNA, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setrxant"
	},
	{ 
	WLAN_SUBCMD_SETTXANTENNA, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE, 
	"settxant"
	},
	{
	WLANSETAUTHALG,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"authalgs",
	},
	{
	WLANSET8021XAUTHALG,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"8021xauthalgs",
	},
	{
	WLANSETENCRYPTIONMODE,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"encryptionmode",
	},
	{ 
	WLANSETREGION,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setregioncode"
	},
	{ 
	WLAN_SET_LISTEN_INTERVAL,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setlisteninter"
	},
	{ 
	WLAN_SET_MULTIPLE_DTIM,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setmultipledtim"
	},
	{ 
	WLAN_SET_ATIM_WINDOW,
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"atimwindow"
	},
	{ 
	WLANSETBCNAVG, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setbcnavg"
	},
	{ 
	WLANSETDATAAVG, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	IW_PRIV_TYPE_NONE,
	"setdataavg"
	},
	{ 
	WLAN_SETNONE_GETONEINT, 
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	""
	},
	{ 
	WLANGETREGION,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	"getregioncode"
	},
	{ 
	WLAN_GET_LISTEN_INTERVAL,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	"getlisteninter"
	},
	{ 
	WLAN_GET_MULTIPLE_DTIM,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	"getmultipledtim"
	},
	{ 
	WLANGETBCNAVG,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	"getbcnavg"
	},
	{ 
	WLAN_SETNONE_GETTWELVE_CHAR,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_CHAR | 12,
	""
	},
	{ 
	WLAN_SUBCMD_GETRXANTENNA, 
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_CHAR | 12, 
	"getrxant"
	},
	{ 
	WLAN_SUBCMD_GETTXANTENNA, 
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_CHAR | 12, 
	"gettxant"
	},
	{
	WLANDEEPSLEEP,
	IW_PRIV_TYPE_CHAR | 1, 
	IW_PRIV_TYPE_CHAR | 6, 
	"deepsleep"
	},

	{
	WLANHOSTSLEEPCFG,
	IW_PRIV_TYPE_CHAR | 15,
	IW_PRIV_TYPE_NONE,
	"hostsleepcfg"
	},

	{
	WLAN_SETNONE_GETNONE,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	""
	},
	{
	WLANDEAUTH,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	"deauth"
	},
	{
	WLANADHOCSTOP,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	"adhocstop"
	},
	{
	WLANRADIOON,		
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	"radioon"
	},
	{
	WLANRADIOOFF,		
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	"radiooff"
	},
	{
	WLANREMOVEADHOCAES,		
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_NONE,
	"rmaeskey"
	},
	{
	WLAN_SETNONE_GETBYTE,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE,
	""
	},
#ifdef REASSOCIATION
	{
	WLANREASSOCIATIONAUTO,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE,
	"reasso-on"
	},
	{
	WLANREASSOCIATIONUSER,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE,
	"reasso-off"
	},
#endif /* REASSOCIATION */
	{
	WLANWLANIDLEON,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE,
	"wlanidle-on"
	},
	{
	WLANWLANIDLEOFF,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_BYTE,
	"wlanidle-off"
	},
	{
	WLAN_SET64CHAR_GET64CHAR,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	""
	},
	{
	WLANSLEEPPARAMS,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"sleepparams"
	},

	{
	WLAN_BCA_TIMESHARE,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"bca-ts"
	},
	{
	WLANSCAN_MODE,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"scanmode"
	},
	{ 
	WLAN_GET_ADHOC_STATUS,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"getadhocstatus"
	},
#if defined(IWGENIE_SUPPORT) && (WIRELESS_EXT < 18)
	{
    WLAN_SET_GEN_IE,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"setgenie"
	},
    {
	WLAN_GET_GEN_IE,
	IW_PRIV_TYPE_CHAR | 64,
	IW_PRIV_TYPE_CHAR | 64,
	"getgenie"
	},
#endif
	{
	WLAN_SETWORDCHAR_GETNONE,
	IW_PRIV_TYPE_CHAR | 32,
	IW_PRIV_TYPE_NONE,
	""
	},
	{
	WLANSETADHOCAES,
	IW_PRIV_TYPE_CHAR | 32,
	IW_PRIV_TYPE_NONE,
	"setaeskey"
	},
	{
	WLAN_SETNONE_GETWORDCHAR,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_CHAR | 128,
	""
	},
	{
	WLANGETADHOCAES,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_CHAR | 128,
	"getaeskey"
	},
	{
	WLANVERSION,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_CHAR | 128,
	"version"
	},
	{
	WLANSETWPAIE,
	IW_PRIV_TYPE_CHAR | IW_PRIV_SIZE_FIXED | 24,
	IW_PRIV_TYPE_NONE,
	"setwpaie"
	},
	{ 
	WLAN_SETTENCHAR_GETNONE,
	IW_PRIV_TYPE_CHAR | 10, 
	IW_PRIV_TYPE_NONE,
	""
	},
	{ 
	WLAN_SETNONE_GETTENCHAR,
	IW_PRIV_TYPE_NONE, 
	IW_PRIV_TYPE_CHAR | 10, 
	""
	},
	{
	WLANGETLOG,
	IW_PRIV_TYPE_NONE,
	IW_PRIV_TYPE_CHAR | GETLOG_BUFSIZE,
	"getlog"
	},
        {
        WLAN_SET_GET_SIXTEEN_INT,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
        ""
        },
        {
        WLAN_TPCCFG,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
        "tpccfg"
        },
        {
        WLAN_POWERCFG,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
        "powercfg"
        },
	{
	WLAN_AUTO_FREQ_SET,
	IW_PRIV_TYPE_INT | 16,
	IW_PRIV_TYPE_INT | 16,
	"setafc"
	},
	{
	WLAN_AUTO_FREQ_GET,
	IW_PRIV_TYPE_INT | 16,
	IW_PRIV_TYPE_INT | 16,
	"getafc"
	},
	{
	WLAN_SCANPROBES,
	IW_PRIV_TYPE_INT | 16,
	IW_PRIV_TYPE_INT | 16,
	"scanprobes"
	},
        {
        WLAN_LED_GPIO_CTRL,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
        "ledgpio"
        },
	{ 
	WLAN_SLEEP_PERIOD,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
	"sleeppd"
	},
	{ 
	WLAN_ADAPT_RATESET,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
	"rateadapt"
	},

	{ 
	WLAN_INACTIVITY_TIMEOUT,
        IW_PRIV_TYPE_INT | 16,
        IW_PRIV_TYPE_INT | 16,
	"inactivityto"
	},
	{ 
	WLANSNR, 
	IW_PRIV_TYPE_INT | 16, 
	IW_PRIV_TYPE_INT | 16, 
	"getSNR"
	},
	{ 
	WLAN_GET_RATE,
	IW_PRIV_TYPE_INT | 16, 
	IW_PRIV_TYPE_INT | 16, 
	"getrate"
	},
#if defined(ENABLE_USER_JOIN) || defined(ENABLE_USER_SCAN)
	{ 
        WLAN_SET_GET_2K, /* IOCTL : 14 */
		IW_PRIV_TYPE_BYTE | 2000,
		IW_PRIV_TYPE_BYTE | 2000,
        ""
	},
#endif
	{
		WLAN_SET_USER_SCAN,
		IW_PRIV_TYPE_BYTE | 2000,
		IW_PRIV_TYPE_BYTE | 2000,
		"setuserscan"
	},
	{
		WLAN_GET_SCAN_TABLE,
		IW_PRIV_TYPE_BYTE | 2000,
		IW_PRIV_TYPE_BYTE | 2000,
		"getscantable"
	},

};

struct iw_handler_def wlan_handler_def = {
	num_standard:sizeof(wlan_handler) / sizeof(iw_handler),
	num_private:sizeof(wlan_private_handler) / sizeof(iw_handler),
	num_private_args:sizeof(wlan_private_args) /
	sizeof(struct iw_priv_args),
	standard:(iw_handler *) wlan_handler,
	private:(iw_handler *) wlan_private_handler,
	private_args:(struct iw_priv_args *) wlan_private_args,
};

#ifdef STDCMD
/** 
 *  @brief wlan stdcmd ioctl handler
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param req		        A pointer to ifreq structure
 *  @param cmd			command 
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_stdcmd_ioctl(struct net_device *dev, struct ifreq *req, int cmd)
{
	u8		*tempResponseBuffer;
	CmdCtrlNode	*pCmdNode;
	HostCmd_DS_GEN	gencmd, *pCmdPtr;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	u16		wait_option = 0;
	ENTER();

	/*
	 * Get a free command control node 
	 */
	if (!(pCmdNode = GetFreeCmdCtrlNode(priv))) {
		PRINTM(INFO, "Failed GetFreeCmdCtrlNode\n");
		return -ENOMEM;
	}

	if (!(tempResponseBuffer = kmalloc(3000, GFP_KERNEL))) {
		PRINTM(INFO, "ERROR: Failed to allocate response buffer!\n");
		return -ENOMEM;
	}

	wait_option |= HostCmd_OPTION_WAITFORRSP;
	
	SetCmdCtrlNode(priv, pCmdNode, 0, wait_option, NULL);

	pCmdPtr = (HostCmd_DS_GEN *) pCmdNode->BufVirtualAddr;

	/*
	 * Read to header to get the size 
	 */
	copy_from_user(&gencmd, req->ifr_data, sizeof(gencmd));

	/*
	 * Copy the whole command into the command buffer 
	 */
	copy_from_user(pCmdPtr, req->ifr_data, gencmd.Size);

	pCmdNode->pdata_buf = tempResponseBuffer;
	pCmdNode->CmdFlags |= CMD_F_STDCMD;

	pCmdPtr->SeqNum = ++priv->adapter->SeqNum;
	pCmdPtr->Result = 0;

	PRINTM(INFO, "STDCMD Command: 0x%04x Size: %d SeqNum: %d\n",
			pCmdPtr->Command, pCmdPtr->Size, pCmdPtr->SeqNum);
	HEXDUMP("Command Data", (u8 *) (pCmdPtr), MIN(32, pCmdPtr->Size));
	PRINTM(INFO, "Copying data from : (user)0x%p -> 0x%p(driver)\n",
			req->ifr_data, pCmdPtr);

	QueueCmd(Adapter, pCmdNode, TRUE);
	wake_up_interruptible(&priv->MainThread.waitQ);

	if (wait_option & HostCmd_OPTION_WAITFORRSP) {
		/* Sleep until response is generated by FW */
		wait_event_interruptible(pCmdNode->cmdwait_q, 
						pCmdNode->CmdWaitQWoken);
	}
		
	/* Copy the response back to user space */
	pCmdPtr = (HostCmd_DS_GEN *) tempResponseBuffer;

	if (copy_to_user(req->ifr_data, tempResponseBuffer, pCmdPtr->Size))
		PRINTM(INFO, "ERROR: copy_to_user failed!\n");

	kfree(tempResponseBuffer);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

#endif	/* STDCMD */
/** 
 *  @brief Get SNR 
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param wreq		        A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_snr(wlan_private *priv, struct iwreq *wrq)
{
  	int ret = WLAN_STATUS_SUCCESS;
  	wlan_adapter *Adapter = priv->adapter;
  	int data[4];

  	ENTER();
  	memset(data,0,sizeof(data));
  	if(wrq->u.data.length)
		copy_from_user(data,wrq->u.data.pointer,MIN(wrq->u.data.length,4) * sizeof(int));
  	if ((wrq->u.data.length == 0) || (data[0] == 0) || (data[0] == 1))
  	{ 
    		ret = PrepareAndSendCommand(priv, 
          		HostCmd_CMD_802_11_RSSI,
          		0, HostCmd_OPTION_WAITFORRSP,
          		0, NULL);

    		if (ret) 
    		{
      			LEAVE();
      			return ret;
    		}
  	}

  	if (wrq->u.data.length == 0)
  	{
    		data[0] = Adapter->SNR[TYPE_BEACON][TYPE_NOAVG];
    		data[1] = Adapter->SNR[TYPE_BEACON][TYPE_AVG];
      		data[2] = Adapter->SNR[TYPE_RXPD][TYPE_NOAVG];
    		data[3] = Adapter->SNR[TYPE_RXPD][TYPE_AVG] / AVG_SCALE;
    		copy_to_user(wrq->u.data.pointer,data,sizeof(int) * 4);
    		wrq->u.data.length = 4;
  	}
  	else if (data[0] == 0)
  	{
    		data[0] = Adapter->SNR[TYPE_BEACON][TYPE_NOAVG];
    		copy_to_user(wrq->u.data.pointer,data,sizeof(int));
    		wrq->u.data.length = 1;
  	}
  	else if (data[0] == 1)
  	{
    		data[0] = Adapter->SNR[TYPE_BEACON][TYPE_AVG];
    		copy_to_user(wrq->u.data.pointer,data,sizeof(int));
    		wrq->u.data.length = 1;
  	}
  	else if (data[0] == 2)
  	{
      		data[0] = Adapter->SNR[TYPE_RXPD][TYPE_NOAVG];
   		copy_to_user(wrq->u.data.pointer,data,sizeof(int));
    		wrq->u.data.length = 1;
  	}
  	else if (data[0] == 3)
  	{
    		data[0] = Adapter->SNR[TYPE_RXPD][TYPE_AVG] / AVG_SCALE;
    		copy_to_user(wrq->u.data.pointer,data,sizeof(int));
    		wrq->u.data.length = 1;
  	}
  	else
  	{
    		return -ENOTSUPP;
  	}
    
  	LEAVE();
  	return WLAN_STATUS_SUCCESS;
}




/** 
 *  @brief Get RSSI 
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param wreq		        A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_rssi(wlan_private *priv, struct iwreq *wrq)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_adapter	*Adapter = priv->adapter;
	int		temp;
	int data = 0;
	
	ENTER();
			
	if(wrq->u.data.length) {
		copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
		if ((data == 0) || (data == 1)) { 
			ret = PrepareAndSendCommand(priv, 
					HostCmd_CMD_802_11_RSSI,
					0,HostCmd_OPTION_WAITFORRSP,
					0, NULL);

			if (ret) {
				LEAVE();
				return ret;
			}
		}

		switch (data) {
		case 0:
		temp = CAL_RSSI(Adapter->SNR[TYPE_BEACON][TYPE_NOAVG],
					Adapter->NF[TYPE_BEACON][TYPE_NOAVG]);
				break;
		case 1:
		temp = CAL_RSSI(Adapter->SNR[TYPE_BEACON][TYPE_AVG],
				Adapter->NF[TYPE_BEACON][TYPE_AVG]);
				break;
		case 2:
		temp = CAL_RSSI(Adapter->SNR[TYPE_RXPD][TYPE_NOAVG],
					Adapter->NF[TYPE_RXPD][TYPE_NOAVG]);
				break;
		case 3:
		temp = CAL_RSSI(Adapter->SNR[TYPE_RXPD][TYPE_AVG] / AVG_SCALE,
				Adapter->NF[TYPE_RXPD][TYPE_AVG] / AVG_SCALE);
				break;
		default:
			return -ENOTSUPP;
		}
		
		if (copy_to_user(wrq->u.data.pointer, 
					&temp, sizeof(int))) {
			PRINTM(INFO, "copy_to_user failed!\n");
			return -EFAULT;
		}
		wrq->u.data.length = 1;
	} else
		return -EINVAL;
 
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get NF
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param wreq		        A pointer to iwreq structure 
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_nf(wlan_private *priv, struct iwreq *wrq)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_adapter	*Adapter = priv->adapter;
	int		temp;
	int data = 0;

	ENTER();

	if(wrq->u.data.length) {
		copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
		if ((data == 0) || (data == 1) ) { 
			ret = PrepareAndSendCommand(priv, 
					HostCmd_CMD_802_11_RSSI,
					0,HostCmd_OPTION_WAITFORRSP,
					0, NULL);

			if (ret) {
				LEAVE();
				return ret;
			}
		}

		switch (data) {
			case 0:
				temp = Adapter->NF[TYPE_BEACON][TYPE_NOAVG];
				break;
			case 1:
				temp = Adapter->NF[TYPE_BEACON][TYPE_AVG];
				break;
			case 2:
				temp = Adapter->NF[TYPE_RXPD][TYPE_NOAVG];
				break;
			case 3:
				temp = Adapter->
					NF[TYPE_RXPD][TYPE_AVG] / AVG_SCALE;
				break;
			default:
				return -ENOTSUPP;
		}

		temp = CAL_NF(temp);
		
		PRINTM(INFO, "***temp = %d\n", temp);		
		if (copy_to_user(wrq->u.data.pointer, 
					&temp, sizeof(int))) {
			PRINTM(INFO, "copy_to_user failed!\n");
			return -EFAULT;
		}
		wrq->u.data.length = 1;
	} else
		return -EINVAL;
		
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Remove AES key
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_remove_aes(wlan_private *priv)
{
	wlan_adapter	*Adapter = priv->adapter;	
	WLAN_802_11_KEY	key;
	int 	ret = WLAN_STATUS_SUCCESS;

	ENTER();

	if(Adapter->InfrastructureMode != Wlan802_11IBSS ||
			Adapter->MediaConnectStatus == 
					WlanMediaStateConnected)
		return -EOPNOTSUPP;
					
	Adapter->AdhocAESEnabled = FALSE;

	memset(&key, 0, sizeof(WLAN_802_11_KEY));
	PRINTM(INFO, "WPA2: DISABLE AES_KEY\n");
	key.KeyLength = WPA_AES_KEY_LEN;
	key.KeyIndex = 0x40000000;

	ret = PrepareAndSendCommand(priv,
		HostCmd_CMD_802_11_KEY_MATERIAL,
		HostCmd_ACT_SET,
		HostCmd_OPTION_WAITFORRSP,
		!(KEY_INFO_ENABLED), &key);
	
	LEAVE();
	
	return ret;
}


/** 
 *  @brief Get Support Rates
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_getrate_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	wlan_adapter		*Adapter = priv->adapter;
	WLAN_802_11_RATES	rates;
	int rate[16];
	int			i;

	ENTER();

	memset(rates, 0, sizeof(rates));
	memset(rate,0,sizeof(rate));
	wrq->u.data.length = get_active_data_rates(Adapter, rates);
	if(wrq->u.data.length > 16)
		wrq->u.data.length = 16;

	for (i = 0; i < wrq->u.data.length; i++)
	{
		rates[i] &= ~0x80;
		rate[i] = rates[i];
	}

	if (copy_to_user(wrq->u.data.pointer, rate, wrq->u.data.length * sizeof(int))) {
		PRINTM(INFO, "Copy to user failed\n");
		return -EFAULT;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get Adhoc Status
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_get_adhoc_status_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	char 		status[64];
	wlan_adapter	*Adapter = priv->adapter;
	
	memset(status, 0, sizeof(status));

	switch (Adapter->InfrastructureMode) {
	case Wlan802_11IBSS:
		if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
			if (Adapter->AdhocCreate == TRUE)
				memcpy(&status, "AdhocStarted", sizeof(status));
			else
				memcpy(&status, "AdhocJoined", sizeof(status));
		} else {
			memcpy(&status, "AdhocIdle", sizeof(status));
		}
		break;
	case Wlan802_11Infrastructure:	
		memcpy(&status, "InfraMode", sizeof(status));
		break;
	default:	
		memcpy(&status, "AutoUnknownMode", sizeof(status));
		break;
	}

	PRINTM(INFO, "Status = %s\n", status);
	wrq->u.data.length = strlen(status) + 1;

	if (wrq->u.data.pointer) {
		if (copy_to_user(wrq->u.data.pointer, 
					&status, wrq->u.data.length))
			return -EFAULT;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get Driver Version
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_version_ioctl(wlan_private *priv, struct ifreq *req)
{
	int	len;
	char	buf[128];
	struct iwreq	*wrq = (struct iwreq *) req;

	ENTER();

	get_version(priv->adapter, buf, sizeof(buf) - 1);

	len = strlen(buf);
	if (wrq->u.data.pointer) {
		if(copy_to_user(wrq->u.data.pointer, buf, len)){
			PRINTM(INFO, "CopyToUser failed\n");
			return -EFAULT;
		}
		wrq->u.data.length = len;
	}

	PRINTM(INFO, "wlan version: %s\n", buf);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Read/Write adapter registers
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_regrdwr_ioctl(wlan_private *priv, struct ifreq *req)
{
	wlan_ioctl_regrdwr	regrdwr;
	wlan_offset_value	offval;
	wlan_adapter		*Adapter = priv->adapter;
	int			ret = WLAN_STATUS_SUCCESS;

	ENTER();

	if(copy_from_user(&regrdwr, req->ifr_data,	sizeof(regrdwr))){
		PRINTM(INFO, "copy of regrdwr for wlan_regrdwr_ioctl from user failed \n");
		LEAVE();
 		return -EFAULT;	
	}

	if(regrdwr.WhichReg == REG_EEPROM) {
		PRINTM(MSG, "Inside RDEEPROM\n");
		Adapter->pRdeeprom = (char *) kmalloc((regrdwr.NOB + sizeof(regrdwr)), GFP_KERNEL);
		if(!Adapter -> pRdeeprom)
			return -ENOMEM;
		memcpy(Adapter->pRdeeprom, &regrdwr,sizeof(regrdwr));
		/* +14 is for Action, Offset, and NOB in 
		 * response */
		PRINTM(INFO, "Action:%d Offset: %x NOB: %02x\n",
					regrdwr.Action, 
					regrdwr.Offset, regrdwr.NOB);
		
		ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_EEPROM_ACCESS,
				regrdwr.Action,HostCmd_OPTION_WAITFORRSP,
				0, &regrdwr);
	
		if (ret) {
			if(Adapter -> pRdeeprom)
				kfree(Adapter -> pRdeeprom);
			LEAVE();
			return ret;
		}

		mdelay(10);

		/*
		 * Return the result back to the user 
		 */

		if (regrdwr.Action == HostCmd_ACT_GEN_READ) {
			if(copy_to_user(req->ifr_data, Adapter -> pRdeeprom, sizeof(regrdwr) + regrdwr.NOB))
			{
				if(Adapter -> pRdeeprom)
					kfree(Adapter -> pRdeeprom);
				PRINTM(INFO, "copy of regrdwr for wlan_regrdwr_ioctl to user failed \n");
				LEAVE();
				return -EFAULT;	
			}
		}
			
		if(Adapter -> pRdeeprom)
			kfree(Adapter -> pRdeeprom);

		return WLAN_STATUS_SUCCESS;
	}

	offval.offset = regrdwr.Offset;
	offval.value = (regrdwr.Action) ? regrdwr.Value : 0x00;

	PRINTM(INFO, "RegAccess: %02x Action:%d "
			"Offset: %04x Value: %04x\n",
			regrdwr.WhichReg, regrdwr.Action, 
			offval.offset, offval.value);

	/*
	 * regrdwr.WhichReg should contain the command that
	 * corresponds to which register access is to be 
	 * performed HostCmd_CMD_MAC_REG_ACCESS 0x0019
	 * HostCmd_CMD_BBP_REG_ACCESS 0x001a 
	 * HostCmd_CMD_RF_REG_ACCESS 0x001b 
	 */
	ret = PrepareAndSendCommand(priv, regrdwr.WhichReg, 
			regrdwr.Action,HostCmd_OPTION_WAITFORRSP,
			0, &offval);

	if (ret) {
		LEAVE();
		return ret;
	}

	mdelay(10);

	/*
	 * Return the result back to the user 
	 */
	regrdwr.Value = Adapter->OffsetValue.value;
	if (regrdwr.Action == HostCmd_ACT_GEN_READ) {
		if(copy_to_user(req->ifr_data, &regrdwr, sizeof(regrdwr))){
			PRINTM(INFO, "copy of regrdwr for wlan_regrdwr_ioctl to user failed \n");
			LEAVE();
 			return -EFAULT;	
		}
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}
		
#if (defined SD) || (defined SPI)
/** 
 *  @brief Cmd52 read/write register
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_cmd52rdwr_ioctl(wlan_private *priv, struct ifreq *req)
{
	u8 buf[7];
	u8 rw, func, dat = 0xff;
	u32 reg;

	ENTER();

	copy_from_user(buf, req->ifr_data, sizeof(buf));

	rw = buf[0];
	func = buf[1];
	reg = buf[5];
	reg = (reg << 8) + buf[4];
	reg = (reg << 8) + buf[3];
	reg = (reg << 8) + buf[2];

	if(rw != 0)
		dat = buf[6];
	
	PRINTM(INFO, "rw=%d func=%d reg=0x%08X dat=0x%02X\n",rw,func,	reg,dat);

	if(rw == 0) {
		if (sbi_read_ioreg(priv, func, reg, &dat) < 0) {
			PRINTM(INFO, "sdio_read_ioreg: reading register 0x%X failed\n"
									,reg);
			dat = 0xff;
		}
	}
	else {
		if (sbi_write_ioreg(priv, func, reg, dat) < 0) {
			PRINTM(INFO, "sdio_read_ioreg: writing register 0x%X failed\n"
									,reg);
			dat = 0xff;
		}
	}
	copy_to_user(req->ifr_data, &dat, sizeof(dat));

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Cmd53 read/write register
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_cmd53rdwr_ioctl(wlan_private *priv, struct ifreq *req)
{
	return -EINVAL;
}
#endif



/** 
 *  @brief Convert ascii string to Hex integer
 *     
 *  @param d                    A pointer to integer buf
 *  @param s			A pointer to ascii string 
 *  @param dlen			the length o fascii string
 *  @return 	   	        number of integer  
 */
static int ascii2hex(u8 *d, char *s, u32 dlen)
{
	int i;
	u8 n;
    
	memset(d, 0x00, dlen);

	for (i = 0; i < dlen * 2; i++) {
		if ((s[i] >= 48) && (s[i] <= 57))
			n = s[i] - 48;
		else if ((s[i] >= 65) && (s[i] <= 70))
			n = s[i] - 55;
		else if ((s[i] >= 97) && (s[i] <= 102))
			n = s[i] - 87;
		else
			break;
		if ((i % 2) == 0)
			n = n * 16;
		d[i / 2] += n;
	}

	return i;
}

/** 
 *  @brief Set adhoc aes key
 *   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_setadhocaes_ioctl(wlan_private *priv, struct ifreq *req)
{
	u8  key_ascii[32];
	u8  key_hex[16];
	int ret = 0;
	struct iwreq *wrq = (struct iwreq *) req;
	wlan_adapter *Adapter = priv->adapter;

	WLAN_802_11_KEY		key;

	ENTER();

	if (Adapter->InfrastructureMode != Wlan802_11IBSS)
		return -EOPNOTSUPP;

	if (Adapter->MediaConnectStatus == WlanMediaStateConnected)
		return -EOPNOTSUPP;

	if(copy_from_user(key_ascii, wrq->u.data.pointer, sizeof(key_ascii))){
		PRINTM(INFO, "wlan_setadhocaes_ioctl copy from user failed \n");
		LEAVE();
 		return -EFAULT;	
	}
  
	Adapter->AdhocAESEnabled = TRUE;
	ascii2hex(key_hex, key_ascii, sizeof(key_hex));

	HEXDUMP("wlan_setadhocaes_ioctl", key_hex, sizeof(key_hex));

	PRINTM(INFO, "WPA2: ENABLE AES_KEY\n");
	key.KeyLength = WPA_AES_KEY_LEN;
	key.KeyIndex = 0x40000000;
	memcpy(key.KeyMaterial, key_hex, key.KeyLength );

	ret = PrepareAndSendCommand(priv,
		HostCmd_CMD_802_11_KEY_MATERIAL,
		HostCmd_ACT_SET,
	 	HostCmd_OPTION_WAITFORRSP,
 		KEY_INFO_ENABLED, &key);

	LEAVE();
	return ret;
}

/** 
 *  @brief Get adhoc aes key   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_getadhocaes_ioctl(wlan_private *priv, struct ifreq *req)
{
	u8 *tmp;
	u8 key_ascii[33];
	u8 key_hex[16];
	int i, ret = 0;
	struct iwreq *wrq = (struct iwreq *) req;
	wlan_adapter *Adapter = priv->adapter;
	WLAN_802_11_KEY		key;

	ENTER();

	memset(key_hex, 0x00, sizeof(key_hex));

	PRINTM(INFO, "WPA2: ENABLE AES_KEY\n");
	key.KeyLength = WPA_AES_KEY_LEN;
	key.KeyIndex = 0x40000000;
	memcpy(key.KeyMaterial, key_hex, key.KeyLength );

	ret = PrepareAndSendCommand(priv,
		HostCmd_CMD_802_11_KEY_MATERIAL,
		HostCmd_ACT_GET,
		HostCmd_OPTION_WAITFORRSP,
 		KEY_INFO_ENABLED, &key);

	if (ret) {
		LEAVE();
		return ret;
	}

	memcpy(key_hex, Adapter->aeskey.KeyParamSet.Key, sizeof(key_hex));

	HEXDUMP("wlan_getadhocaes_ioctl", key_hex, sizeof(key_hex));

	wrq->u.data.length = sizeof(key_ascii) + 1;

	memset(key_ascii, 0x00, sizeof(key_ascii));
	tmp = key_ascii;
            
	for (i = 0; i < sizeof(key_hex); i++) 
		tmp += sprintf(tmp, "%02x", key_hex[i]);

	if (wrq->u.data.pointer) {
		if (copy_to_user(wrq->u.data.pointer, &key_ascii, 
					sizeof(key_ascii))) {
			PRINTM(INFO, "copy_to_user failed\n");
			return -EFAULT;
		}
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Set/Get WPA IE   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_setwpaie_ioctl(wlan_private *priv, struct ifreq *req)
{
	struct iwreq	*wrq = (struct iwreq *) req;
	wlan_adapter	*Adapter = priv->adapter;
	int		ret = WLAN_STATUS_SUCCESS;

	ENTER();

	if (wrq->u.data.length) {
		if (wrq->u.data.length > sizeof(Adapter->Wpa_ie)){
			PRINTM(INFO, "failed to copy WPA IE, too big \n");
			return -EFAULT;
		}
		if (copy_from_user(Adapter->Wpa_ie, wrq->u.data.pointer,
							wrq->u.data.length)) {
			PRINTM(INFO, "failed to copy WPA IE \n");
			return -EFAULT;
		}
		Adapter->Wpa_ie_len = wrq->u.data.length;
		PRINTM(INFO, "Set Wpa_ie_len=%d IE=%#x\n", Adapter->Wpa_ie_len, 
							Adapter->Wpa_ie[0]);
		HEXDUMP("Wpa_ie", Adapter->Wpa_ie, Adapter->Wpa_ie_len);
		if (Adapter->Wpa_ie[0] == WPA_IE)
			Adapter->SecInfo.WPAEnabled = TRUE;
		else if (Adapter->Wpa_ie[0] == WPA2_IE)
			Adapter->SecInfo.WPA2Enabled = TRUE;
		else {
			Adapter->SecInfo.WPAEnabled = FALSE;
			Adapter->SecInfo.WPA2Enabled = FALSE;
		}
	} else {
		memset(Adapter->Wpa_ie, 0, sizeof(Adapter->Wpa_ie));
		Adapter->Wpa_ie_len = wrq->u.data.length;
		PRINTM(INFO, "Reset Wpa_ie_len=%d IE=%#x\n", Adapter->Wpa_ie_len,
							Adapter->Wpa_ie[0]);
		Adapter->SecInfo.WPAEnabled = FALSE;
		Adapter->SecInfo.WPA2Enabled = FALSE;
	}

	// enable/disable RSN in firmware if WPA is enabled/disabled
	// depending on variable Adapter->SecInfo.WPAEnabled is set or not
	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_ENABLE_RSN,
			HostCmd_ACT_SET, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	LEAVE();
	return ret;
}

/** 
 *  @brief Set atim window   
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_atim_window_ioctl(wlan_private *priv, struct ifreq *req)
{
	wlan_adapter	*Adapter = priv->adapter;
	struct	ifreq	*treq;

	ENTER();
	
	PRINTM(INFO, "WLAN_SET_ATIM_WINDOW\n");
	treq = (struct ifreq *)((u8 *)req + 4);
	Adapter->AtimWindow = (int)treq->ifr_data;
	Adapter->AtimWindow = MIN(Adapter->AtimWindow, 50);
	PRINTM(INFO, "AtimWindow is set to %d\n", Adapter->AtimWindow);

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Region  
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_setregion_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		region_code;
	struct	ifreq	*treq;

	ENTER();

	treq = (struct ifreq *)((u8 *)req + 4);
	region_code = (int) treq->ifr_data;

	if (wlan_set_region(priv, (u16) region_code))
		return -EFAULT;
	
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Set Listen Interval
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_listen_interval_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		interval;
	struct	ifreq	*treq;

	ENTER();
					
	treq = (struct ifreq *)((u8 *)req + 4);
	interval = (int) treq->ifr_data;
					
	priv->adapter->ListenInterval = (u16) interval;
		
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set RX antenna
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_subcmd_setrxantenna_ioctl(wlan_private *priv, struct ifreq *req)
{
	struct	ifreq	*treq;
	PRINTM(INFO, "WLAN_SUBCMD_SETRXANTENNA\n");
	treq = (struct ifreq *)((u8 *)req + 4);
	return SetRxAntenna(priv, (int)treq->ifr_data);
}


/** 
 *  @brief Set TX antenna
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_subcmd_settxantenna_ioctl(wlan_private *priv, struct ifreq *req)
{
	struct	ifreq	*treq;
	PRINTM(INFO, "WLAN_SUBCMD_SETTXANTENNA\n");
	treq = (struct ifreq *)((u8 *)req + 4);
	return SetTxAntenna(priv, (int)treq->ifr_data);
}

/** 
 *  @brief Set multiple dtim
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set_multiple_dtim_ioctl(wlan_private *priv, struct ifreq *req)
{
	struct	ifreq	*treq;
	u32		mdtim;
	int		ret = -EINVAL;

	ENTER();
	
	treq = (struct ifreq *)((u8 *)req + 4);
	if (treq) {
		mdtim = (u32)treq->ifr_data;
		if (((mdtim >= MRVDRV_MIN_MULTIPLE_DTIM) &&
			(mdtim <= MRVDRV_MAX_MULTIPLE_DTIM))
			|| (mdtim == MRVDRV_IGNORE_MULTIPLE_DTIM)
		) {
			priv->adapter->MultipleDtim = mdtim;
			ret = WLAN_STATUS_SUCCESS;
		}
	}

	if (ret)
		PRINTM(INFO, "Invalid parameter, MultipleDtim not changed.\n");

	LEAVE();
	return ret;
}

/** 
 *  @brief Set authentication mode
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_setauthalg_ioctl(wlan_private *priv, struct ifreq *req)
{
	int 	alg;
	struct	ifreq	*treq;
	struct iwreq	*wrq = (struct iwreq *) req;
	wlan_adapter	*Adapter = priv->adapter;
	
	ENTER();

	if (wrq->u.data.flags == 0) {
		//from iwpriv subcmd
		treq = (struct ifreq *)((u8 *)req + 4);
		alg = (int)treq->ifr_data;
	} else {	
		//from wpa_supplicant subcmd
		copy_from_user(&alg,wrq->u.data.pointer,sizeof(alg));
	}
	
	PRINTM(INFO, "auth alg is %#x\n",alg);

	switch (alg) {
		case AUTH_ALG_SHARED_KEY:
			Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeShared;
			break;
		case AUTH_ALG_NETWORK_EAP:
			Adapter->SecInfo.AuthenticationMode = 
						Wlan802_11AuthModeNetworkEAP;
			break;
		case AUTH_ALG_OPEN_SYSTEM:
		default:
			Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeOpen;
			break;
	}
	
	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set 802.1x authentication mode
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_set8021xauthalg_ioctl(wlan_private *priv, struct ifreq *req)
{
	int 		alg;
	struct	ifreq	*treq;
	struct iwreq	*wrq = (struct iwreq *) req;
	
	ENTER();
	
	if (wrq->u.data.flags == 0) {
		//from iwpriv subcmd
		treq = (struct ifreq *)((u8 *)req + 4);
		alg = (int)treq->ifr_data;
	} else {	
		//from wpa_supplicant subcmd
		copy_from_user(&alg,wrq->u.data.pointer,sizeof(int));
	}
	PRINTM(INFO, "802.1x auth alg is %#x\n",alg);
	priv->adapter->SecInfo.Auth1XAlg = alg;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Encryption mode
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_setencryptionmode_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		mode;
	struct	ifreq	*treq;
	struct iwreq	*wrq = (struct iwreq *) req;

	ENTER();
	
	if (wrq->u.data.flags == 0) {
		//from iwpriv subcmd
		treq = (struct ifreq *)((u8 *)req + 4);
		mode = (int)treq->ifr_data;
	}
	else {	
		//from wpa_supplicant subcmd
		copy_from_user(&mode,wrq->u.data.pointer,sizeof(int));
	}
	PRINTM(INFO, "encryption mode is %#x\n",mode);
	priv->adapter->SecInfo.EncryptionMode = mode;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get Rx antenna
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_subcmd_getrxantenna_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		len;
	char		buf[8];
	struct iwreq	*wrq = (struct iwreq *) req;

	ENTER();

	PRINTM(INFO, "WLAN_SUBCMD_GETRXANTENNA\n");
	len = GetRxAntenna(priv, buf);

	wrq->u.data.length = len;
	if (wrq->u.data.pointer) {
		if (copy_to_user(wrq->u.data.pointer, &buf, len)) {
			PRINTM(INFO, "CopyToUser failed\n");
			return -EFAULT;
		}
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get Tx antenna
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_subcmd_gettxantenna_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		len;
	char		buf[8];
	struct iwreq	*wrq = (struct iwreq *) req;

	ENTER();

	PRINTM(INFO, "WLAN_SUBCMD_GETTXANTENNA\n");
	len = GetTxAntenna(priv, buf);

	wrq->u.data.length = len;
	if (wrq->u.data.pointer) {
		if (copy_to_user(wrq->u.data.pointer, &buf, len)) {
			PRINTM(INFO, "CopyToUser failed\n");
			return -EFAULT;
		}
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get/Set DeepSleep mode
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_deepsleep_ioctl(wlan_private *priv, struct ifreq *req)
{
	int		ret = WLAN_STATUS_SUCCESS;
	char		status[128];
	struct iwreq	*wrq = (struct iwreq *) req;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();
	
	// check if it is in IEEE PS mode or not
	if (Adapter->PSMode != Wlan802_11PowerModeCAM) {
		PRINTM(MSG, "Cannot process Deep Sleep command in IEEE Power"
							" Save mode.\n");
		return -EINVAL;
	}

	if (*(char *)req->ifr_data == '0') {
		PRINTM(INFO, "Exit Deep Sleep Mode.\n");
		sprintf(status, "setting to off ");
		SetDeepSleep(priv, FALSE);	
	} else if (*(char *)req->ifr_data == '1') {
		PRINTM(INFO, "Enter Deep Sleep Mode.\n");
		sprintf(status, "setting to on ");
		SetDeepSleep(priv, TRUE);
	} else if (*(char *)req->ifr_data == '2') {
		PRINTM(INFO, "Get Deep Sleep Mode.\n");
		if (Adapter->IsDeepSleep == TRUE) {
			sprintf(status, "on ");
		} else {
			sprintf(status, "off ");
		}
	} else {
		PRINTM(INFO, "unknown option = %d\n",*(u8 *)req->ifr_data);
		return  -EINVAL;
	}

	if (wrq->u.data.pointer) {
		if (copy_to_user(wrq->u.data.pointer, &status, strlen(status)))
			return -EFAULT;
		wrq->u.data.length = strlen(status);
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Config hostsleep parameter
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_hostsleepcfg_ioctl(wlan_private *priv, struct iwreq *wrq) 
{
	wlan_adapter	*Adapter = priv->adapter;
	char		buf[16];
	int		ret = WLAN_STATUS_SUCCESS;
	int 		gpio, gap;
	

	memset(buf, 0, sizeof(buf));
	if (copy_from_user(buf, wrq->u.data.pointer, 
				MIN(sizeof(buf)-1,
				wrq->u.data.length))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}
	buf[15] = 0;

	sscanf(buf,"%x %x %x",&Adapter->HostSleepCfgParam.conditions,
		&gpio, &gap );
	Adapter->HostSleepCfgParam.gpio = (u8)gpio;
	Adapter->HostSleepCfgParam.gap	 = (u8)gap;

	PRINTM(MSG, "conditions=0x%x gpio=%d gap=%d\n",
		Adapter->HostSleepCfgParam.conditions,
		Adapter->HostSleepCfgParam.gpio,
		Adapter->HostSleepCfgParam.gap);

	ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_HOST_SLEEP_CFG,
				0, HostCmd_OPTION_WAITFORRSP,
				0, &Adapter->HostSleepCfgParam);

	return ret;
}

/** 
 *  @brief subscribe event
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_subscribe_event(wlan_private *priv, struct ifreq *req)
{
	EventSubscribe		*Events;
	wlan_adapter		*Adapter = priv->adapter;
	int			ret;
	
	ENTER();

	Events = kmalloc(sizeof(EventSubscribe), GFP_KERNEL);
	if(!Events){
		PRINTM(MSG, "kmalloc no memory !!\n");
		return -ENOMEM;
		
	}
	
	if (copy_from_user(Events, req->ifr_data + SKIP_CMDNUM,
			       sizeof(EventSubscribe))) {
			PRINTM(INFO, "Copy from user failed\n");
			kfree(Events);
			return -EFAULT;
	}
	HEXDUMP("SubscribeEvents: ", req->ifr_data + SKIP_CMDNUM,
					sizeof(EventSubscribe));

	ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_SUBSCRIBE_EVENT,
				0,HostCmd_OPTION_WAITFORRSP,
			       	0, Events);

	if (Events->Action == HostCmd_ACT_GET) {
		PRINTM(INFO, "Suscribed Events = 0x%X\n", Adapter->GetEvents);
		if (copy_to_user(req->ifr_data + SKIP_CMDNUM, 
					&Adapter->GetEvents, sizeof(u16))) {
			PRINTM(INFO, "Copy to user Failed\n");
			kfree(Events);
			return -EFAULT;
		}
	}

	kfree(Events);

	LEAVE();
	return ret;
}


/** 
 *  @brief Get/Set Cal data ext
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_caldata_ext_ioctl(wlan_private *priv, struct ifreq *req)
{
	int			ret = WLAN_STATUS_SUCCESS;
	u8			buf[8];
	HostCmd_DS_802_11_CAL_DATA_EXT	CalData;
	
	ENTER();
	
	memset(&CalData, 0, sizeof(HostCmd_DS_802_11_CAL_DATA_EXT));
			
	if (copy_from_user(&CalData.Action, req->ifr_data + SKIP_CMDNUM,
						sizeof(HostCmd_DS_802_11_CAL_DATA_EXT))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}

	HEXDUMP("Cal data ext", (u8 *)&CalData.Action, 
					sizeof(HostCmd_DS_802_11_CAL_DATA_EXT));

	PRINTM(INFO, "CalData.Action = 0x%0X\n", CalData.Action);
	
	switch (CalData.Action) {
	case WLANSET_CAL_DATA_EXT:
		ret = PrepareAndSendCommand(priv, 
					HostCmd_CMD_802_11_CAL_DATA_EXT,
					HostCmd_ACT_GEN_SET,	
					HostCmd_OPTION_WAITFORRSP,
					0, &CalData);

		if (ret) {
			LEAVE();
			return ret;
		}
					
		if (copy_to_user(req->ifr_data, &priv->adapter->caldataext,
					sizeof(HostCmd_DS_802_11_CAL_DATA_EXT) + 
					sizeof(HostCmd_DS_GEN))) {
			PRINTM(INFO, "Copy from user failed\n");
			return -EFAULT;
		}

		buf[0] = 0x01;
		buf[1] = 0x00;
		break;
			
	case WLANGET_CAL_DATA_EXT:
		ret = PrepareAndSendCommand(priv, 
					HostCmd_CMD_802_11_CAL_DATA_EXT,	
					HostCmd_ACT_GEN_GET,
					HostCmd_OPTION_WAITFORRSP,
					0, &CalData);

		if (ret) {
			LEAVE();
			return ret;
		}

		if (copy_to_user(req->ifr_data, &priv->adapter->caldataext,
					sizeof(HostCmd_DS_802_11_CAL_DATA_EXT) +
					sizeof(HostCmd_DS_GEN))) {
			PRINTM(INFO, "Copy from user failed\n");
			return -EFAULT;
		}
			break;
	}

	LEAVE();
	return ret;
}


/** 
 *  @brief Get/Set sleep period 
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_sleep_period(wlan_private *priv, struct iwreq *wrq)
{
	int ret;		
	int data;
	wlan_adapter *Adapter = priv->adapter;
	HostCmd_DS_802_11_SLEEP_PERIOD	sleeppd;
       
	ENTER();

	if (wrq->u.data.length > 1)
		return -ENOTSUPP;
	
	memset(&sleeppd, 0, sizeof(sleeppd));
	memset(&Adapter->sleep_period, 0, sizeof(SleepPeriod));

        if (wrq->u.data.length == 0) {
		sleeppd.Action = wlan_cpu_to_le16(HostCmd_ACT_GEN_GET);
        } else {
		copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
		/* sleep period is between 0 or 10 ... 60 */
		if ((data <= 60 && data >= 10) || (data == 0)){
			sleeppd.Action = 
				wlan_cpu_to_le16(HostCmd_ACT_GEN_SET);
			sleeppd.Period = data;
		} else
			return -EINVAL;
        }

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SLEEP_PERIOD,
			0, HostCmd_OPTION_WAITFORRSP,
			0, (void *) &sleeppd);
	
	data = (int) Adapter->sleep_period.period;
	copy_to_user(wrq->u.data.pointer,&data,sizeof(int));
        wrq->u.data.length = 1;

	LEAVE();
	return ret;
}

/** 
 *  @brief Get/Set adapt rate 
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_adapt_rateset(wlan_private *priv, struct iwreq *wrq)
{
	int			ret;
	wlan_adapter		*Adapter = priv->adapter;
	int data[2];

	ENTER();
	memset(data,0,sizeof(data));
	if (!wrq->u.data.length) {
		PRINTM(INFO, "Get ADAPT RATE SET\n");
		ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_RATE_ADAPT_RATESET,
				HostCmd_ACT_GEN_GET, HostCmd_OPTION_WAITFORRSP,
			      	0, NULL);
		data[0] = Adapter->EnableHwAuto;
		data[1] = Adapter->RateBitmap;
		copy_to_user(wrq->u.data.pointer,data,sizeof(int) * 2);
#define GET_TWO_INT	2
		wrq->u.data.length = GET_TWO_INT;
	} else {
		PRINTM(INFO, "Set ADAPT RATE SET\n");
		if(wrq->u.data.length > 2)
			return -EINVAL;
		copy_from_user(data,wrq->u.data.pointer,sizeof(int) * wrq->u.data.length);
		Adapter->EnableHwAuto = data[0];
		Adapter->RateBitmap = data[1];
		ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_RATE_ADAPT_RATESET,
				HostCmd_ACT_GEN_SET,HostCmd_OPTION_WAITFORRSP,
			      	0, NULL);
	}

	LEAVE();
	return ret;	
}

/** 
 *  @brief Get/Set inactivity timeout
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_inactivity_timeout(wlan_private *priv, struct iwreq *wrq)
{
	int ret;		
    	int data = 0;
	u16		timeout = 0;
	ENTER();
	if (wrq->u.data.length > 1)	
		return -ENOTSUPP;
	
        if (wrq->u.data.length == 0) {
		/* Get*/
		ret = PrepareAndSendCommand(priv,
				HostCmd_CMD_802_11_INACTIVITY_TIMEOUT,
				HostCmd_ACT_GET, HostCmd_OPTION_WAITFORRSP,
				0, &timeout);
		data = timeout;
		copy_to_user(wrq->u.data.pointer,&data,sizeof(int));
        } else {
		/* Set */
		copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
		timeout =data;
		ret = PrepareAndSendCommand(priv, 
				HostCmd_CMD_802_11_INACTIVITY_TIMEOUT,
				HostCmd_ACT_SET, HostCmd_OPTION_WAITFORRSP,
				0, &timeout);
	}

        wrq->u.data.length = 1;

	LEAVE();
	return ret;
}


/** 
 *  @brief Get LOG
 *  @param priv                 A pointer to wlan_private structure
 *  @param wrq			A pointer to iwreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_getlog_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	int		ret;
	char		buf[GETLOG_BUFSIZE-1];
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	PRINTM(INFO, " GET STATS\n"); 

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_GET_LOG,
			0, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	if (ret) {
		LEAVE();
		return ret;
	}

	if (wrq->u.data.pointer) {
		sprintf(buf, "\n  mcasttxframe %u failed %u retry %u " 
				"multiretry %u framedup %u "
				"rtssuccess %u rtsfailure %u ackfailure %u\n"
				"rxfrag %u mcastrxframe %u fcserror %u "
				"txframe %u wepundecryptable %u ",
				Adapter->LogMsg.mcasttxframe,
				Adapter->LogMsg.failed,
				Adapter->LogMsg.retry,
				Adapter->LogMsg.multiretry,
				Adapter->LogMsg.framedup,
				Adapter->LogMsg.rtssuccess,
				Adapter->LogMsg.rtsfailure,
				Adapter->LogMsg.ackfailure,
				Adapter->LogMsg.rxfrag,
				Adapter->LogMsg.mcastrxframe,
				Adapter->LogMsg.fcserror,
				Adapter->LogMsg.txframe,
				Adapter->LogMsg.wepundecryptable);
		wrq->u.data.length = strlen(buf) + 1;
		copy_to_user(wrq->u.data.pointer, buf, wrq->u.data.length);
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief Read/Write CF register
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_regcfrdwr_ioctl(wlan_private *priv, struct ifreq *req)
{
	wlan_ioctl_cfregrdwr regrdwr;

	ENTER();

	copy_from_user(&regrdwr, req->ifr_data, 
					sizeof(regrdwr));

	if (regrdwr.Action) {
		PRINTM(INFO, "Write value into CF register\n");
		sbi_write_cfreg(priv, regrdwr.Offset, regrdwr.Value);
	} else {
		PRINTM(INFO, "Read value from CF register\n");
		regrdwr.Value = sbi_read_cfreg(priv, 
		regrdwr.Offset);
		copy_to_user(req->ifr_data, &regrdwr, 
						sizeof(regrdwr));
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief config sleep parameters
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_sleep_params_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	int ret;
	wlan_adapter 	*Adapter = priv->adapter;
	wlan_ioctl_sleep_params_config  sp;

	ENTER();

	memset(&sp, 0, sizeof(sp));
	
	if (copy_from_user(&sp, wrq->u.data.pointer,
				MIN(sizeof(sp), wrq->u.data.length))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}
	
	memcpy(&Adapter->sp, &sp.Error, sizeof(SleepParams));
	
	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_SLEEP_PARAMS,
			sp.Action, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	if (!ret && !sp.Action && wrq->u.data.pointer) {
		memcpy(&sp.Error, &Adapter->sp, sizeof(SleepParams));
		copy_to_user(wrq->u.data.pointer, &sp, sizeof(sp));
		wrq->u.data.length = sizeof(sp);
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Read the CIS Table
 *  @param priv                 A pointer to wlan_private structure
 *  @param req			A pointer to ifreq structure
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
static int wlan_do_getcis_ioctl(wlan_private *priv, struct ifreq *req)
{
	int 		ret = WLAN_STATUS_SUCCESS;
	struct iwreq	*wrq = (struct iwreq *) req;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if (wrq->u.data.pointer) {
		copy_to_user(wrq->u.data.pointer, Adapter->CisInfoBuf, 
							Adapter->CisInfoLen);
		wrq->u.data.length = Adapter->CisInfoLen;
	}

	LEAVE();
	return ret;
}


/** 
 *  @brief Set BCA timeshare
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_bca_timeshare_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	int 				ret;
	wlan_adapter 			*Adapter = priv->adapter;
	wlan_ioctl_bca_timeshare_config bca_ts;

	ENTER();

	memset(&bca_ts, 0, sizeof(HostCmd_DS_802_11_BCA_TIMESHARE));
	
	if (copy_from_user(&bca_ts, wrq->u.data.pointer,
				MIN(sizeof(bca_ts), wrq->u.data.length))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}

	PRINTM(INFO, "TrafficType=%x TimeShareInterva=%x BTTime=%x\n", 
			bca_ts.TrafficType, bca_ts.TimeShareInterval, 
			bca_ts.BTTime);

	ret = PrepareAndSendCommand(priv, 
			HostCmd_CMD_802_11_BCA_CONFIG_TIMESHARE,
			bca_ts.Action, HostCmd_OPTION_WAITFORRSP,
			0, &bca_ts);

	if (!ret && !bca_ts.Action && wrq->u.data.pointer) {
		copy_to_user(wrq->u.data.pointer, &Adapter->bca_ts, 
								sizeof(bca_ts));
		wrq->u.data.length = sizeof(HostCmd_DS_802_11_BCA_TIMESHARE);
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Set scan type
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_scan_type_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	u8	buf[12];
	u8	*option[] = { "active", "passive", "get", };
	int	i, max_options = (sizeof(option)/sizeof(option[0]));
	int	ret = WLAN_STATUS_SUCCESS;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if( wlan_get_state_11d( priv) == ENABLE_11D ) {
		PRINTM(INFO, "11D: Cannot set scantype when 11D enabled\n");
		return -EFAULT;
	}

	memset(buf, 0, sizeof(buf));

	if (copy_from_user(buf, wrq->u.data.pointer, MIN(sizeof(buf), 
					wrq->u.data.length))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}


	PRINTM(INFO, "Scan Type Option = %s\n", buf);

	buf[sizeof(buf)-1] = '\0';

	for (i = 0; i < max_options; i++) {
		if (!strcmp(buf, option[i]))
			break;
	}

	switch (i) {
		case 0:
			Adapter->ScanType = HostCmd_SCAN_TYPE_ACTIVE;
			break;
		case 1:
			Adapter->ScanType = HostCmd_SCAN_TYPE_PASSIVE;
			break;
		case 2:
			wrq->u.data.length = 
				strlen(option[Adapter->ScanType]) + 1;

			if (copy_to_user(wrq->u.data.pointer, 
						option[Adapter->ScanType],
						wrq->u.data.length)) {
				PRINTM(INFO, "Copy to user failed\n");
				ret = -EFAULT;
			}

			break;
		default:
			PRINTM(INFO, "Invalid Scan Type Ioctl Option\n");
			ret = -EINVAL;
			break;
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Set scan mode
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_scan_mode_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	wlan_adapter *Adapter = priv->adapter;
	u8	buf[12];
	u8	*option[] = { "bss", "ibss", "any","get"};
	int	i, max_options = (sizeof(option)/sizeof(option[0]));
	int	ret = WLAN_STATUS_SUCCESS;

	ENTER();

	memset(buf, 0, sizeof(buf));

	if (copy_from_user(buf, wrq->u.data.pointer, MIN(sizeof(buf), 
					wrq->u.data.length))) {
		PRINTM(INFO, "Copy from user failed\n");
		return -EFAULT;
	}

	PRINTM(INFO, "Scan Mode Option = %s\n", buf);

	buf[sizeof(buf)-1] = '\0';

	for (i = 0; i < max_options; i++) {
		if (!strcmp(buf, option[i]))
			break;
	}

	switch (i) {

		case 0:
			Adapter->ScanMode = HostCmd_BSS_TYPE_BSS;
			break;
		case 1:
			Adapter->ScanMode = HostCmd_BSS_TYPE_IBSS;
			break;
		case 2:
			Adapter->ScanMode = HostCmd_BSS_TYPE_ANY;
			break;
		case 3:

			wrq->u.data.length = 
				strlen(option[Adapter->ScanMode-1]) + 1;

			PRINTM(INFO, "Get Scan Mode Option = %s\n", 
					option[Adapter->ScanMode-1]);

			PRINTM(INFO, "Scan Mode Length %d\n", wrq->u.data.length);

			if (copy_to_user(wrq->u.data.pointer, 
						option[Adapter->ScanMode-1],
							wrq->u.data.length)) {
				PRINTM(INFO, "Copy to user failed\n");
				ret = -EFAULT;
			}
			PRINTM(INFO, "GET Scan Type Option after copy = %s\n", 
					(char *)wrq->u.data.pointer);

			break;

		default:
			PRINTM(INFO, "Invalid Scan Mode Ioctl Option\n");
			ret = -EINVAL;
			break;
	}

	LEAVE();
	return ret;
}

/** 
 *  @brief Get/Set Adhoc G Rate
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_do_set_grate_ioctl(wlan_private *priv, struct iwreq *wrq)
{
	wlan_adapter	*Adapter = priv->adapter;
	int 		data,data1;
	
	ENTER();
    	if(wrq->u.data.length == 0)
		return -EINVAL;
	copy_from_user(&data1,wrq->u.data.pointer,sizeof(int));
	switch (data1) {
		case 0:
			Adapter->adhoc_grate_enabled = FALSE;
			break;
		case 1:
			Adapter->adhoc_grate_enabled = TRUE;
			break;
		case 2:
			data = Adapter->adhoc_grate_enabled;
			copy_to_user(wrq->u.data.pointer, &data, sizeof(int) );
			wrq->u.data.length = 1;
			break;
		default:
			return -EINVAL;
	}

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Get/Set Firmware wakeup method
 *  
 *  @param priv		A pointer to wlan_private structure
 *  @param wrq	   	A pointer to user data
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
static int wlan_cmd_fw_wakeup_method( wlan_private *priv, struct iwreq *wrq )
{
	wlan_adapter	*Adapter = priv->adapter;
	u16		action;
	u16 		method;
	int		ret;
	int 		data;
	
	ENTER();

	if (wrq->u.data.length == 0 || !wrq->u.data.pointer) {
		action = HostCmd_ACT_GET;
	}
	else {
		action = HostCmd_ACT_SET;
		copy_from_user(&data, wrq->u.data.pointer,sizeof(int));
		switch (data ) {
		case 0:
			Adapter->fwWakeupMethod = WAKEUP_FW_UNCHANGED;
			break;
		case 1:
			Adapter->fwWakeupMethod = WAKEUP_FW_THRU_INTERFACE;
			break;
		case 2:
			Adapter->fwWakeupMethod = WAKEUP_FW_THRU_GPIO;
			break;
		default:
			return -EINVAL;
		}
	}

	ret = PrepareAndSendCommand(priv,
		HostCmd_CMD_802_11_FW_WAKEUP_METHOD, action,
		HostCmd_OPTION_WAITFORRSP, 0,
		&Adapter->fwWakeupMethod);

	if (action == HostCmd_ACT_GET) {
		method = Adapter->fwWakeupMethod;
		copy_to_user(wrq->u.data.pointer, &method, sizeof(method));
		wrq->u.data.length = 1;
	}

	LEAVE();
	return ret;
}


/********************************************************
		Global Functions
********************************************************/
/** 
 *  @brief ioctl function - entry point
 *  
 *  @param dev		A pointer to net_device structure
 *  @param req	   	A pointer to ifreq structure
 *  @param cmd 		command
 *  @return 	   	WLAN_STATUS_SUCCESS--success, otherwise fail
 */
int wlan_do_ioctl(struct net_device *dev, struct ifreq *req, int cmd)
{
	int		subcmd = 0;
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	struct iwreq	*wrq = (struct iwreq *) req;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)
			&& (cmd != WLANDEEPSLEEP)
			/* here we exclude all sub-cmds under SET1INT_GET1INT cmd
			   in order to get SDIOCLOCK sub-cmd to work; need a fix later */
			&& (cmd != WLAN_SETONEINT_GETONEINT)
		) {
		PRINTM(MSG, "():%s IOCTLS called when station is"
					" in DeepSleep\n",__FUNCTION__);
		return -EBUSY;
	}

	PRINTM(INFO, "wlan_do_ioctl: ioctl cmd = 0x%x\n", cmd);
	switch (cmd) {
	case WLANEXTSCAN: 
		ret = wlan_extscan_ioctl(priv, req);
		break;
#ifdef STDCMD
	case WLANSTDCMD:
		ret = wlan_stdcmd_ioctl(dev, req, cmd);
		break;
#endif

	case WLANCISDUMP:	/* Read CIS Table  */
		ret = wlan_do_getcis_ioctl(priv, req);
		break;
		
	case WLANSCAN_TYPE :
		PRINTM(INFO, "Scan Type Ioctl\n");
		ret = wlan_scan_type_ioctl(priv, wrq);
		break;

#ifdef MANF_CMD_SUPPORT
	case WLANMANFCMD:
		PRINTM(INFO, "Entering the Manufacturing ioctl SIOCCFMFG\n");
		ret = wlan_mfg_command(priv, (void *) req->ifr_data);
		
		PRINTM(INFO, "Manufacturing Ioctl %s\n", (ret) ? "failed":"success");
		break;
#endif

	case WLANREGRDWR:	/* Register read write command */
		ret = wlan_regrdwr_ioctl(priv, req);
		break;

#if (defined SD) || (defined SPI)
	case WLANCMD52RDWR:	/* CMD52 read/write command */
		ret = wlan_cmd52rdwr_ioctl(priv, req);
		break;

	case WLANCMD53RDWR:	/* CMD53 read/write command */
		ret = wlan_cmd53rdwr_ioctl(priv, req);
		break;
#endif

	case SIOCSIWENCODE:	/* set encoding token & mode for WPA */
		ret = wlan_set_encode(dev, NULL, &(wrq->u.data),
							wrq->u.data.pointer);
		break;
	case WLAN_SETNONE_GETNONE : /* set WPA mode on/off ioctl #20 */
		switch(wrq->u.data.flags) {
		case WLANDEAUTH :
			PRINTM(INFO, "Deauth\n");
			wlan_send_deauth(priv);
			break;
			
		case WLANADHOCSTOP:
			PRINTM(INFO, "Adhoc stop\n");
			ret = wlan_do_adhocstop_ioctl(priv);
			break;

		case WLANRADIOON:
			wlan_radio_ioctl(priv, RADIO_ON);
			break;
			
		case WLANRADIOOFF:
			wlan_radio_ioctl(priv, RADIO_OFF);
			break;
		case WLANREMOVEADHOCAES:
			ret = wlan_remove_aes(priv);
			break;
		}	/* End of switch */
		break;

	case WLAN_SETWORDCHAR_GETNONE :
		switch(wrq->u.data.flags) {
		case WLANSETADHOCAES:
			ret = wlan_setadhocaes_ioctl(priv, req);
			break;
		}
		break;

	case WLAN_SETNONE_GETWORDCHAR:
		switch(wrq->u.data.flags) {
		case WLANGETADHOCAES:
			ret = wlan_getadhocaes_ioctl(priv, req);
			break;
		case WLANVERSION:	/* Get driver version */
			ret = wlan_version_ioctl(priv, req);
			break;
		}
		break;
	case WLANSETWPAIE:
		ret = wlan_setwpaie_ioctl(priv, req);
		break;

	case WLAN_SETONEINT_GETONEINT:
		switch(wrq->u.data.flags) {
		case WLANNF :		   
			ret  = wlan_get_nf(priv, wrq);
			break;
				
		case WLANRSSI :		
			ret = wlan_get_rssi(priv, wrq);
			break;
		case WLANWMM_ENABLE:
			ret = wlan_wmm_enable_ioctl(priv, wrq);
			break;

		case WLANBGSCAN:
			if (wrq->u.data.length) {
				int data,data1;
				copy_from_user(&data1,wrq->u.data.pointer,sizeof(int));
				switch (data1) {
					case CMD_DISABLED:
						PRINTM(INFO, "Background scan is set to disable\n");
						ret = wlan_bg_scan_enable(priv, FALSE);
						break;
					case CMD_ENABLED:
						PRINTM(INFO, "Background scan is set to enable\n");
						ret = wlan_bg_scan_enable(priv, TRUE);
						break;
					case CMD_GET:
						data = (Adapter->bgScanConfig->Enable == TRUE) ? 
							CMD_ENABLED : CMD_DISABLED;
						wrq->u.data.length = 1;
						copy_to_user(wrq->u.data.pointer, &data, sizeof(int));
						break;
					default:			
						ret = -EINVAL;
						PRINTM(INFO, "Background scan: wrong parameter\n");
						break;
				}
			}
			else {
				ret = -EINVAL;
				PRINTM(INFO, "Background scan: wrong parameter\n");
			}
			break;

		case WLANENABLE11D:
			ret = wlan_cmd_enable_11d( priv, wrq );
			break;

		case WLANADHOCGRATE:
			ret = wlan_do_set_grate_ioctl(priv, wrq);
			break;


		case WLAN_WMM_QOSINFO:
			{
				int data;
				if (wrq->u.data.length == 1)
				{
					copy_from_user(&data, wrq->u.data.pointer,sizeof(int));
					Adapter->wmm.qosinfo = (u8)data; 
				}
				else
				{
					data = (int) Adapter->wmm.qosinfo;
					copy_to_user(wrq->u.data.pointer,&data,sizeof(int));
					wrq->u.data.length = 1;
				}
			}
			break;
		case WLAN_LISTENINTRVL:
			if(!wrq->u.data.length) {				
				int data;
				PRINTM(INFO, "Get LocalListenInterval Value\n");
#define GET_ONE_INT	1				
				data = Adapter->LocalListenInterval;
				copy_to_user(wrq->u.data.pointer,&data,sizeof(int));
				wrq->u.data.length = GET_ONE_INT;
			} else {
				int data;
				copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
				PRINTM(INFO, "Set LocalListenInterval = %d\n",data);
#define MAX_U16_VAL	65535
				if(data > MAX_U16_VAL) {
					PRINTM(INFO, "Exceeds U16 value\n");
					return -EINVAL;
				}
				Adapter->LocalListenInterval = data;
			}
			break;
		case WLAN_FW_WAKEUP_METHOD:
			ret = wlan_cmd_fw_wakeup_method( priv, wrq );
			break;		
		case WLAN_TXCONTROL:
			ret = wlan_txcontrol(priv, wrq ); //adds for txcontrol ioctl
			break;


		case  WLANNULLGEN:
			ret = wlan_null_pkg_gen(priv, wrq );
			/* enable/disable null pkg generation*/
			break;
		case WLAN_NULLPKTINTERVAL:
			ret = wlan_null_pkt_interval(priv, wrq);
			break;
		default:
			ret = -EOPNOTSUPP;
			break;
		}
		break;	
	
	case WLAN_SETONEINT_GETNONE:
		/* The first 4 bytes of req->ifr_data is sub-ioctl number
		 * after 4 bytes sits the payload.
		 */
		subcmd = wrq->u.data.flags;	//from wpa_supplicant subcmd
		
		if(!subcmd)
			subcmd = (int)req->ifr_data;	//from iwpriv subcmd
		
		switch (subcmd) {
		case WLAN_SUBCMD_SETRXANTENNA : /* SETRXANTENNA */
			ret = wlan_subcmd_setrxantenna_ioctl(priv, req);
			break;
				
		case WLAN_SUBCMD_SETTXANTENNA : /* SETTXANTENNA */
			ret = wlan_subcmd_settxantenna_ioctl(priv, req);
			break;
		case WLAN_SET_ATIM_WINDOW:
			ret = wlan_set_atim_window_ioctl(priv, req);
			break;
		case WLANSETBCNAVG:
			priv->adapter->bcn_avg_factor = 
				(int) ((struct ifreq *)((u8 *)req + 4))->ifr_data;
			if(priv->adapter->bcn_avg_factor == 0)
				priv->adapter->bcn_avg_factor = DEFAULT_BCN_AVG_FACTOR;
			if(priv->adapter->bcn_avg_factor > DEFAULT_BCN_AVG_FACTOR)
				priv->adapter->bcn_avg_factor = DEFAULT_BCN_AVG_FACTOR;
			ret = WLAN_STATUS_SUCCESS;
			break;
		case WLANSETDATAAVG:
			priv->adapter->data_avg_factor = 
				(int) ((struct ifreq *)((u8 *)req + 4))->ifr_data;
			if(priv->adapter->data_avg_factor == 0)
				priv->adapter->data_avg_factor = DEFAULT_DATA_AVG_FACTOR;
			if(priv->adapter->data_avg_factor > DEFAULT_DATA_AVG_FACTOR)
				priv->adapter->data_avg_factor = DEFAULT_DATA_AVG_FACTOR;
			ret = WLAN_STATUS_SUCCESS;
			break;
		case WLANSETREGION:
			ret = wlan_setregion_ioctl(priv, req);
			break;
				
		case WLAN_SET_LISTEN_INTERVAL:
			ret = wlan_set_listen_interval_ioctl(priv, req);
			break;
				
		case WLAN_SET_MULTIPLE_DTIM:
			ret = wlan_set_multiple_dtim_ioctl(priv, req);
			break;

		case WLANSETAUTHALG:
			ret = wlan_setauthalg_ioctl(priv, req);
			break;
		
		case WLANSET8021XAUTHALG:
			ret = wlan_set8021xauthalg_ioctl(priv, req);
			break;

		case WLANSETENCRYPTIONMODE:
			ret = wlan_setencryptionmode_ioctl(priv, req);
			break;
		default:
			ret = -EOPNOTSUPP;
			break;
		}
		
		break;

	case WLAN_SETNONE_GETTWELVE_CHAR: /* Get Antenna settings */
		/* 
		 * We've not used IW_PRIV_TYPE_FIXED so sub-ioctl number is
		 * in flags of iwreq structure, otherwise it will be in
		 * mode member of iwreq structure.
		 */
		switch ((int)wrq->u.data.flags) {
		case WLAN_SUBCMD_GETRXANTENNA: /* Get Rx Antenna */
			ret = wlan_subcmd_getrxantenna_ioctl(priv, req);
			break;

		case WLAN_SUBCMD_GETTXANTENNA:  /* Get Tx Antenna */
			ret = wlan_subcmd_gettxantenna_ioctl(priv, req);
			break;
		}
		break;
		
	case WLAN_SETNONE_GETBYTE:
		switch(wrq->u.data.flags) {
#ifdef REASSOCIATION
		case WLANREASSOCIATIONAUTO:
			reassociation_on(priv);
			break;
		case WLANREASSOCIATIONUSER:
			reassociation_off(priv);
			break;
#endif /* REASSOCIATION */
		case WLANWLANIDLEON:
			wlanidle_on(priv);
			break;
		case WLANWLANIDLEOFF:
			wlanidle_off(priv);
			break;
		}	
		
		break;		
		
	case WLANDEEPSLEEP:
		ret = wlan_deepsleep_ioctl(priv, req);
		break;

	case WLANHOSTSLEEPCFG:
		ret = wlan_do_hostsleepcfg_ioctl(priv, wrq); 
		break;
			
	case WLAN_SET64CHAR_GET64CHAR:
		switch ((int)wrq->u.data.flags) {

		case WLANSLEEPPARAMS:
			ret = wlan_sleep_params_ioctl(priv, wrq);
			break;

		case WLAN_BCA_TIMESHARE:
			ret = wlan_bca_timeshare_ioctl(priv, wrq);
			break;
		case WLANSCAN_MODE :
			PRINTM(INFO, "Scan Mode Ioctl\n");
			ret = wlan_scan_mode_ioctl(priv, wrq);
			break;
		
		case WLAN_GET_ADHOC_STATUS:
			ret = wlan_get_adhoc_status_ioctl(priv, wrq);
			break;
#if defined(IWGENIE_SUPPORT) && (WIRELESS_EXT < 18)                    
        	case WLAN_SET_GEN_IE:
            		ret = wlan_set_gen_ie_ioctl(priv, wrq);
            		break;
        	case WLAN_GET_GEN_IE:
            		ret = wlan_get_gen_ie_ioctl(priv, wrq);
            		break;
#endif
        	}
        	break;
	case WLAN_SETCONF_GETCONF:
		PRINTM(INFO, "The WLAN_SETCONF_GETCONF=0x%x is %d\n",WLAN_SETCONF_GETCONF, *(u8 *)req->ifr_data);
		switch (*(u8 *)req->ifr_data) {
		case CAL_DATA_EXT_CONFIG:
		ret = wlan_do_caldata_ext_ioctl(priv, req);
		break;
		case BG_SCAN_CONFIG:
			ret = wlan_do_bg_scan_config_ioctl(priv, req);
			break;

		case SUBSCRIBE_EVENT:
			ret = wlan_do_subscribe_event(priv, req);
			break;

		case WMM_TSPEC:
			ret = wlan_do_wmm_tspec_ioctl(priv, req);
			break;
		case WMM_ACK_POLICY:
			ret = wlan_do_wmm_ack_policy_ioctl(priv, req);
			break;	
		case WMM_PARA_IE:
			ret = wlan_do_wmm_para_ie_ioctl(priv, req);
			break;
		}
		break;

	case WLAN_SETNONE_GETONEINT:
		switch((int)req->ifr_data) {
		case WLANGETBCNAVG:
			req->ifr_data = (char *)((u32)Adapter->bcn_avg_factor);
			break;

		case WLANGETREGION:
			req->ifr_data = (char *)((u32)Adapter->RegionCode);
			break;

		case WLAN_GET_LISTEN_INTERVAL:
			req->ifr_data = (char *)((u32)Adapter->ListenInterval);
			break;

		case WLAN_GET_MULTIPLE_DTIM:
			req->ifr_data = (char *)((u32)Adapter->MultipleDtim);
			break;
		default:
			ret = -EOPNOTSUPP;

		}

		break;
		
	case WLAN_SETTENCHAR_GETNONE:
		switch ((int)wrq->u.data.flags) {

		}
		break;

	case WLAN_SETNONE_GETTENCHAR:
		switch ((int)wrq->u.data.flags) {
		}
		break;

    	case WLANGETLOG:
        	ret = wlan_do_getlog_ioctl(priv, wrq);
        	break;

	case WLANREGCFRDWR:
		ret = wlan_do_regcfrdwr_ioctl(priv, req);
		break;

  	case WLAN_SET_GET_SIXTEEN_INT:
    		switch ((int)wrq->u.data.flags) {
      		case WLAN_TPCCFG:
     	 	{
        		int data[5];
			HostCmd_DS_802_11_TPC_CFG cfg;
        		memset(&cfg, 0, sizeof(cfg));
        		if ((wrq->u.data.length > 1) && (wrq->u.data.length != 5))
          			return WLAN_STATUS_FAILURE;

        		if (wrq->u.data.length == 0){
          			cfg.Action  = wlan_cpu_to_le16(HostCmd_ACT_GEN_GET);
        		}
        		else
        		{
  	  			copy_from_user(data,wrq->u.data.pointer, sizeof(int) * 5);
			        cfg.Action  = wlan_cpu_to_le16(HostCmd_ACT_GEN_SET);
          			cfg.Enable  = data[0];
          			cfg.UseSNR  = data[1];
          			cfg.P0      = data[2];
          			cfg.P1      = data[3];
          			cfg.P2      = data[4];
        		}

        		ret = PrepareAndSendCommand(priv,HostCmd_CMD_802_11_TPC_CFG,0, 
          			HostCmd_OPTION_WAITFORRSP,0,(void *) &cfg);

        		data[0] = cfg.Enable;
        		data[1] = cfg.UseSNR;
        		data[2] = cfg.P0;
        		data[3] = cfg.P1;
        		data[4] = cfg.P2;
			copy_to_user(wrq->u.data.pointer,data,sizeof(int) * 5);
        		wrq->u.data.length = 5;
      		}
      		break;
  
      		case WLAN_POWERCFG:
      		{
      			int data[4];
        		HostCmd_DS_802_11_PWR_CFG cfg;
        		memset(&cfg, 0, sizeof(cfg));
        		if ((wrq->u.data.length > 1) && (wrq->u.data.length != 4))
          			return WLAN_STATUS_FAILURE;
        		if (wrq->u.data.length == 0){
				cfg.Action  = wlan_cpu_to_le16(HostCmd_ACT_GEN_GET);
        		}
        		else
        		{
	  			copy_from_user(data,wrq->u.data.pointer, sizeof(int) * 4);		
          			cfg.Action  = wlan_cpu_to_le16(HostCmd_ACT_GEN_SET);
          			cfg.Enable  = data[0];
          			cfg.PA_P0   = data[1];
          			cfg.PA_P1   = data[2];
          			cfg.PA_P2   = data[3];
        		}
        		ret = PrepareAndSendCommand(priv,HostCmd_CMD_802_11_PWR_CFG,0, 
          			HostCmd_OPTION_WAITFORRSP,0,(void *) &cfg);
        		data[0] = cfg.Enable;
        		data[1] = cfg.PA_P0;
        		data[2] = cfg.PA_P1;
        		data[3] = cfg.PA_P2; 
			copy_to_user(wrq->u.data.pointer,data,sizeof(int) * 4);
        		wrq->u.data.length = 4;
      		}
      		break;
  		case WLAN_AUTO_FREQ_SET: 
      		{
			int data[3];
        		HostCmd_DS_802_11_AFC afc;
        		memset(&afc, 0, sizeof(afc));
			if (wrq->u.data.length != 3)
		 		return WLAN_STATUS_FAILURE;
			copy_from_user(data,wrq->u.data.pointer,sizeof(int)*3);
        		afc.afc_auto    = data[0];

        		if (afc.afc_auto != 0) {
          			afc.afc_thre    = data[1];
          			afc.afc_period  = data[2];
        		} else {
          			afc.afc_toff    = data[1];
          			afc.afc_foff    = data[2];
        		}
        		ret = PrepareAndSendCommand(priv,HostCmd_CMD_802_11_SET_AFC,0, 
				HostCmd_OPTION_WAITFORRSP,0,(void *) &afc);
      		}
      		break;
		case WLAN_AUTO_FREQ_GET: 
      		{
        		int data[3];
        		HostCmd_DS_802_11_AFC afc;
        		memset(&afc, 0, sizeof(afc));
        		ret = PrepareAndSendCommand(priv,HostCmd_CMD_802_11_GET_AFC,0, 
          			HostCmd_OPTION_WAITFORRSP,0,(void *) &afc);
        		data[0] = afc.afc_auto;
        		data[1] = afc.afc_toff;
        		data[2] = afc.afc_foff;
			copy_to_user(wrq->u.data.pointer,data,sizeof(int) * 3);
        		wrq->u.data.length = 3;
      		}
      		break;
		case WLAN_SCANPROBES: 
      		{
			int data;      
        		if (wrq->u.data.length > 0){
	  			copy_from_user(&data,wrq->u.data.pointer,sizeof(int));
          			Adapter->ScanProbes = data;
			}
        		else
			{
          			data = Adapter->ScanProbes;
	  			copy_to_user(wrq->u.data.pointer,&data,sizeof(int));
			}
        		wrq->u.data.length = 1;
      		}
      		break;
    		case WLAN_LED_GPIO_CTRL:
    		{
      			int i;
      			int data[16];
      
      			HostCmd_DS_802_11_LED_CTRL ctrl;
      			MrvlIEtypes_LedGpio_t *gpio = (MrvlIEtypes_LedGpio_t *) ctrl.data;

      			memset(&ctrl, 0, sizeof(ctrl));
        		if (wrq->u.data.length > MAX_LEDS*2)
        			return -ENOTSUPP;
      			if ((wrq->u.data.length % 2) != 0)
        			return -ENOTSUPP;
      			if (wrq->u.data.length == 0) {
        			ctrl.Action = wlan_cpu_to_le16(HostCmd_ACT_GEN_GET);
      			} else {
				copy_from_user(data,wrq->u.data.pointer,sizeof(int)*wrq->u.data.length);
        			ctrl.Action = wlan_cpu_to_le16(HostCmd_ACT_GEN_SET);
        			ctrl.NumLed = wlan_cpu_to_le16(0);
        			gpio->Header.Type = wlan_cpu_to_le16(TLV_TYPE_LED_GPIO);
        			gpio->Header.Len  = wrq->u.data.length;
        			for (i=0; i < wrq->u.data.length; i+=2) {
          				gpio->LedPin[i/2].Led = data[i];
         	 			gpio->LedPin[i/2].Pin = data[i+1];
        			}
      			}
      			ret = PrepareAndSendCommand( priv, HostCmd_CMD_802_11_LED_GPIO_CTRL,0, 
					HostCmd_OPTION_WAITFORRSP,0, (void *) &ctrl);
        		for (i=0; i < gpio->Header.Len; i+=2) {
        			data[i] = gpio->LedPin[i/2].Led;
        			data[i+1] = gpio->LedPin[i/2].Pin;
      			}
      			copy_to_user(wrq->u.data.pointer,data,sizeof(int)* gpio->Header.Len); 	
      			wrq->u.data.length = gpio->Header.Len;
    		}
    		break;
		case WLAN_SLEEP_PERIOD:
			ret = wlan_sleep_period(priv, wrq);
			break;
		case WLAN_ADAPT_RATESET:
			ret = wlan_adapt_rateset(priv, wrq);
			break;
		case WLAN_INACTIVITY_TIMEOUT:
			ret = wlan_inactivity_timeout(priv, wrq); 
			break;
		case WLANSNR :
			ret = wlan_get_snr(priv, wrq);
			break;
		case WLAN_GET_RATE:
			ret = wlan_getrate_ioctl(priv, wrq);
			break;
  		}
  		break;
#if defined(ENABLE_USER_JOIN) || defined(ENABLE_USER_SCAN)
    case WLAN_SET_GET_2K:
        switch ((int)wrq->u.data.flags) {
        case WLAN_SET_USER_SCAN:
            ret = wlan_set_user_scan_ioctl(priv, wrq);
            break;
        case WLAN_GET_SCAN_TABLE:
            ret = wlan_get_scan_table_ioctl(priv, wrq);
            break;
        default:
            ret = -EOPNOTSUPP;
        }
        break;
#endif
	default:
 		ret = -EINVAL;
		break;
	}
	LEAVE();
	return ret;
}


/** 
 *  @brief Get wireless statistics
 *  
 *  @param dev		A pointer to net_device structure
 *  @return 	   	A pointer to iw_statistics buf
 */
struct iw_statistics *wlan_get_wireless_stats(struct net_device *dev)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	int		ret = WLAN_STATUS_SUCCESS;

	ENTER();

	priv->wstats.status = Adapter->InfrastructureMode;
	priv->wstats.discard.retries = priv->stats.tx_errors;

	/* send RSSI command to get beacon RSSI/NF, valid only if associated */
	ret = PrepareAndSendCommand(priv, 
			HostCmd_CMD_802_11_RSSI,
			0, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	priv->wstats.qual.level =
		CAL_RSSI(Adapter->SNR[TYPE_BEACON][TYPE_NOAVG],
		Adapter->NF[TYPE_BEACON][TYPE_NOAVG]);
	priv->wstats.qual.noise = CAL_NF(Adapter->NF[TYPE_BEACON][TYPE_NOAVG]);
	if (Adapter->NF[TYPE_BEACON][TYPE_NOAVG] == 0
		&& Adapter->MediaConnectStatus == WlanMediaStateConnected)
		priv->wstats.qual.noise = MRVDRV_NF_DEFAULT_SCAN_VALUE;
	else
		priv->wstats.qual.noise = 
				CAL_NF(Adapter->NF[TYPE_BEACON][TYPE_NOAVG]);
	priv->wstats.qual.qual = 0;

	PRINTM(INFO, "Signal Level = %#x\n", priv->wstats.qual.level);
	PRINTM(INFO, "Noise = %#x\n", priv->wstats.qual.noise);

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_GET_LOG,
			0, HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	if (!ret) {
		priv->wstats.discard.code = Adapter->LogMsg.wepundecryptable;
		priv->wstats.discard.fragment = Adapter->LogMsg.fcserror;
		priv->wstats.discard.retries = Adapter->LogMsg.retry;
		priv->wstats.discard.misc = Adapter->LogMsg.ackfailure;
	}

	return &priv->wstats;
}

#endif //linux



/** 
 *  @brief Set frequency
 *   
 *  @param priv 		A pointer to wlan_private structure
 *  @param info			A pointer to iw_request_info structure 
 *  @param fwrq			A pointer to iw_freq structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS--success, otherwise--fail
 */
int wlan_set_freq(struct net_device *dev, struct iw_request_info *info, 
		struct iw_freq *fwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	int             rc = -EINPROGRESS;	/* Call commit handler */
	CHANNEL_FREQ_POWER	*cfp;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWFREQ: IOCTLS called when station is "
							"in DeepSleep\n");
		return -EBUSY;
	}


	/*
	 * If setting by frequency, convert to a channel 
	 */
	if (fwrq->e == 1) {

		long	f = fwrq->m / 100000;
		int	c = 0;

		cfp = find_cfp_by_band_and_freq(Adapter, 0, f);
		if (!cfp) {
			PRINTM(INFO, "Invalid freq=%ld\n", f);
			return -EINVAL;
		} 
		
		c = (int) cfp->Channel;

		if (c < 0) 
			return -EINVAL;
	
		fwrq->e = 0;
		fwrq->m = c;
	}

	/*
	 * Setting by channel number 
	 */
	if (fwrq->m > 1000 || fwrq->e > 0) {
		rc = -EOPNOTSUPP;
	} else {
		int	channel = fwrq->m;

		cfp = find_cfp_by_band_and_channel(Adapter, 0, channel);
		if (!cfp) {
			rc = -EINVAL;
		} else {
			if (Adapter->InfrastructureMode == Wlan802_11IBSS) {
				rc = ChangeAdhocChannel(priv, channel);
				/*  If station is WEP enabled, send the 
				 *  command to set WEP in firmware
				 */
				if (Adapter->SecInfo.WEPStatus == 
							Wlan802_11WEPEnabled) {
					PRINTM(INFO, "set_freq: WEP Enabled\n");
					ret = PrepareAndSendCommand(priv,
						HostCmd_CMD_802_11_SET_WEP, 
						0,HostCmd_OPTION_WAITFORRSP
						, OID_802_11_ADD_WEP, 
						NULL);

					if (ret) {
						LEAVE();
						return ret;
					}
						
					Adapter->CurrentPacketFilter |=	
						HostCmd_ACT_MAC_WEP_ENABLE;

					SetMacPacketFilter(priv);
				}
			} else {
				rc = -EOPNOTSUPP;
			}
		}
	}

	LEAVE();
	return rc;
}


/** 
 *  @brief Set Deep Sleep 
 *   
 *  @param adapter 	A pointer to wlan_private structure
 *  @param bDeepSleep	TRUE--enalbe deepsleep, FALSE--disable deepsleep
 *  @return 	   	WLAN_STATUS_SUCCESS-success, otherwise fail
 */

int SetDeepSleep(wlan_private *priv, BOOLEAN bDeepSleep)
{
	int		ret =  WLAN_STATUS_SUCCESS;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	// check if it is in IEEE PS mode or not
	if (Adapter->PSMode != Wlan802_11PowerModeCAM) {
		LEAVE();
		return  WLAN_STATUS_SUCCESS;
	}

	if (bDeepSleep == TRUE) {
		if (Adapter->IsDeepSleep != TRUE) {
			PRINTM(INFO, "Deep Sleep : sleep\n");

			
			// note: the command could be queued and executed later
			//       if there is command in prigressing.
			ret = PrepareAndSendCommand(priv,
					HostCmd_CMD_802_11_DEEP_SLEEP, 0,
					HostCmd_OPTION_WAITFORRSP,
					0, NULL);

			if (ret) {
				LEAVE();
				return ret;
			}
			os_stop_queue(priv);
			os_carrier_off(priv);
		}
	}
	else {
		if (Adapter->IsDeepSleep == TRUE) {
			PRINTM(INFO, "Deep Sleep : wakeup\n");

			if (Adapter->IntCounterSaved)
				Adapter->IntCounter = Adapter->IntCounterSaved;

			if (sbi_exit_deep_sleep(priv))
				PRINTM(INFO, "Deep Sleep : wakeup failed\n");
			

			if (Adapter->IsDeepSleep == TRUE) {
			
				if(interruptible_sleep_on_timeout(&Adapter->ds_awake_q,
						       	WAIT_FOR_SCAN_RRESULT_MAX_TIME) == 0)
					PRINTM(MSG, "ds_awake_q: timer expired\n");
			}

			if(Adapter->IntCounter)
				wake_up_interruptible(&priv->MainThread.waitQ);
		}
	}
	
	LEAVE();
	return  WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief use index to get the data rate
 *   
 *  @param index                The index of data rate
 *  @return 	   		data rate or 0 
 */
u32 index_to_data_rate(u8 index)
{
	if (index >= sizeof(WlanDataRates))
		index = 0;

	return WlanDataRates[index];
}

/** 
 *  @brief use rate to get the index
 *   
 *  @param rate                 data rate
 *  @return 	   		index or 0 
 */
u8 data_rate_to_index(u32 rate)
{
	u8	*ptr;

	if (rate)
		if ((ptr = wlan_memchr(WlanDataRates, (u8)rate, 
						sizeof(WlanDataRates))))
			return (ptr - WlanDataRates);

	return 0;
}


/** 
 *  @brief set data rate
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_set_rate(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	wlan_private		*priv = dev->priv;
	wlan_adapter		*Adapter = priv->adapter;
	u32			data_rate;
	u16			action;
	int 			ret = WLAN_STATUS_SUCCESS;
	WLAN_802_11_RATES	rates;
	u8			*rate;
    
	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWRATE: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}

	PRINTM(INFO, "Vwrq->value = %d\n", vwrq->value);

	if (vwrq->value == -1) {
		action = HostCmd_ACT_SET_TX_AUTO;	// Auto
		Adapter->Is_DataRate_Auto = TRUE;
		Adapter->DataRate = 0;
	} else {
		if (vwrq->value % 100000) {
			return -EINVAL;
		}

		data_rate = vwrq->value / 500000;

		memset(rates, 0, sizeof(rates));
		get_active_data_rates(Adapter, rates);
		rate = rates;
		while (*rate) {
			PRINTM(INFO, "Rate=0x%X  Wanted=0x%X\n",*rate,data_rate);
			if ((*rate & 0x7f) == (data_rate & 0x7f))
				break;
			rate++;
		}
		if (!*rate) {
			PRINTM(MSG, "The fixed data rate 0x%X is out "
					"of range.\n",data_rate);
			return -EINVAL;
		}

		Adapter->DataRate = data_rate;
		action = HostCmd_ACT_SET_TX_FIX_RATE;
		Adapter->Is_DataRate_Auto = FALSE;
	}

	ret = PrepareAndSendCommand(priv, HostCmd_CMD_802_11_DATA_RATE, 
			action,	HostCmd_OPTION_WAITFORRSP,
			0, NULL);

	LEAVE();
	return ret;
}

/** 
 *  @brief get data rate
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_get_rate(struct net_device *dev, struct iw_request_info *info, 
		struct iw_param *vwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();

	if (Adapter->Is_DataRate_Auto) {
		vwrq->fixed = 0;
	} else {
		vwrq->fixed = 1;
	}
	
	vwrq->value = Adapter->DataRate * 500000;

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}


/** 
 *  @brief set wireless mode 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_set_mode(struct net_device *dev,
	      struct iw_request_info *info, u32 * uwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;
	
	WLAN_802_11_NETWORK_INFRASTRUCTURE	WantedMode;

	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWMODE: IOCTLS called when station is in DeepSleep\n");
		return -EBUSY;
	}

	switch (*uwrq) {
		case IW_MODE_ADHOC:
			PRINTM(INFO, "Wanted Mode is ad-hoc: current DataRate=%#x\n",Adapter->DataRate);
			WantedMode = Wlan802_11IBSS;
				Adapter->AdhocChannel = DEFAULT_AD_HOC_CHANNEL;
			break;

		case IW_MODE_INFRA:
			PRINTM(INFO, "Wanted Mode is Infrastructure\n");
			WantedMode = Wlan802_11Infrastructure;
			break;

		case IW_MODE_AUTO:
			PRINTM(INFO, "Wanted Mode is Auto\n");
			WantedMode = Wlan802_11AutoUnknown;
			break;

		default:
			PRINTM(INFO, "Wanted Mode is Unknown: 0x%x\n", *uwrq);
			return -EINVAL;
	}

	if (Adapter->InfrastructureMode == WantedMode ||
			WantedMode == Wlan802_11AutoUnknown) {
		PRINTM(INFO, "Already set to required mode! No change!\n");

		Adapter->InfrastructureMode = WantedMode;

		LEAVE();
		return WLAN_STATUS_SUCCESS;
	}

	if (Adapter->InfrastructureMode == Wlan802_11Infrastructure) {
		if (Adapter->PSState != PS_STATE_FULL_POWER) {
			PSWakeup(priv, HostCmd_OPTION_WAITFORRSP);
		}
		Adapter->PSMode = Wlan802_11PowerModeCAM;
	}
	
	if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		if (Adapter->InfrastructureMode == Wlan802_11Infrastructure) {
			ret = SendDeauthentication(priv);

			if (ret) {
				LEAVE();
				return ret;
			}
		} else if (Adapter->InfrastructureMode == Wlan802_11IBSS) {
			/* If current mode is Adhoc, clean stale information */
			ret = StopAdhocNetwork(priv);

			if (ret) {
				LEAVE();
				return ret;
			}
		}
	} 

	if (Adapter->SecInfo.WEPStatus == Wlan802_11WEPEnabled) {
		/* If there is a key with the specified SSID, 
		 * send REMOVE WEP command, to make sure we clean up
		 * the WEP keys in firmware
		 */
		ret = PrepareAndSendCommand(priv,
			HostCmd_CMD_802_11_SET_WEP,
			0, HostCmd_OPTION_WAITFORRSP
			, OID_802_11_REMOVE_WEP,
			NULL);

		if (ret) {
			LEAVE();
			return ret;
		}

		Adapter->CurrentPacketFilter &= ~HostCmd_ACT_MAC_WEP_ENABLE;

		SetMacPacketFilter(priv);
	}

	Adapter->SecInfo.WEPStatus = Wlan802_11WEPDisabled;
	Adapter->SecInfo.AuthenticationMode = Wlan802_11AuthModeOpen;

	Adapter->InfrastructureMode = WantedMode;

	ret = PrepareAndSendCommand(priv,
			HostCmd_CMD_802_11_SNMP_MIB,
			0, HostCmd_OPTION_WAITFORRSP,
			OID_802_11_INFRASTRUCTURE_MODE,
			NULL);

	if (ret) {
		LEAVE();
		return ret;
	}

	LEAVE();
	return  WLAN_STATUS_SUCCESS;
}

/** 
 *  @brief Set Encryption key
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_set_encode(struct net_device *dev,
		struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{

	PWLAN_802_11_KEY pKey = NULL;

	ENTER();
	
	if (dwrq->length > MAX_KEY_SIZE) {
		pKey = (PWLAN_802_11_KEY) extra;

		if(pKey->KeyLength <= MAX_KEY_SIZE) {
			//dynamic WEP
			dwrq->length = pKey->KeyLength;
			dwrq->flags = pKey->KeyIndex + 1;
			return wlan_set_encode_nonwpa(dev, info, dwrq, 
					pKey->KeyMaterial);
		}
		else {
			//WPA
			return wlan_set_encode_wpa(dev, info, dwrq, extra);
		}
	}
	else {
		//static WEP
		PRINTM(INFO, "Setting WEP\n");
		return wlan_set_encode_nonwpa(dev, info, dwrq, extra);
	}

	return -EINVAL;
}


/** 
 *  @brief set tx power 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_set_txpow(struct net_device *dev, struct iw_request_info *info,
			   		struct iw_param *vwrq, char *extra)
{
	int		ret = WLAN_STATUS_SUCCESS;
	wlan_private *priv= dev->priv;
	wlan_adapter *Adapter= priv->adapter;

	int wlan_radio_ioctl(wlan_private *priv, u8 option);

	u16	dbm;
	
	ENTER();

	if ((Adapter->IsDeepSleep == TRUE)) {
		PRINTM(MSG, "SIOCSIWTXPOW: IOCTLS called when station is"
							" in DeepSleep\n");
		return -EBUSY;
	}
	
	if(vwrq->disabled) {
		wlan_radio_ioctl(priv, RADIO_OFF);
		return WLAN_STATUS_SUCCESS;
	}

	Adapter->Preamble = HostCmd_TYPE_AUTO_PREAMBLE;

	wlan_radio_ioctl(priv, RADIO_ON);

#if WIRELESS_EXT > 14
	if ((vwrq->flags & IW_TXPOW_TYPE) == IW_TXPOW_MWATT) {
		dbm = (u16) mw_to_dbm(vwrq->value);
	}
	else
#endif
		dbm = (u16) vwrq->value;


	/* auto tx power control */

	if (vwrq->fixed == 0)
		dbm = 0xffff;

  	PRINTM(INFO, "<1>TXPOWER SET %d dbm.\n", dbm);

	ret = PrepareAndSendCommand(priv, 
		HostCmd_CMD_802_11_RF_TX_POWER,
		HostCmd_ACT_TX_POWER_OPT_SET_LOW,
		HostCmd_OPTION_WAITFORRSP,
		0, (void *) &dbm);

	LEAVE();
	return ret;
}



/** 
 *  @brief Get current essid 
 *   
 *  @param dev                  A pointer to net_device structure
 *  @param info			A pointer to iw_request_info structure
 *  @param vwrq 		A pointer to iw_param structure
 *  @param extra		A pointer to extra data buf
 *  @return 	   		WLAN_STATUS_SUCCESS --success, otherwise fail
 */
int wlan_get_essid(struct net_device *dev, struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{
	wlan_private	*priv = dev->priv;
	wlan_adapter	*Adapter = priv->adapter;

	ENTER();
	/*
	 * Note : if dwrq->flags != 0, we should get the relevant SSID from
	 * the SSID list... 
	 */

	/*
	 * Get the current SSID 
	 */
	if (Adapter->MediaConnectStatus == WlanMediaStateConnected) {
		memcpy(extra, Adapter->CurBssParams.ssid.Ssid,
				Adapter->CurBssParams.ssid.SsidLength);
		extra[Adapter->CurBssParams.ssid.SsidLength] = '\0';
	} else {
		memset(extra, 0, 32);
		extra[Adapter->CurBssParams.ssid.SsidLength] = '\0';
	}
	/*
	 * If none, we may want to get the one that was set 
	 */

	/* To make the driver backward compatible with WPA supplicant v0.2.4 */
	if (dwrq->length == 32)  /* check with WPA supplicant buffer size */
		dwrq->length = MIN(Adapter->CurBssParams.ssid.SsidLength, 
							IW_ESSID_MAX_SIZE);
	else 		
		dwrq->length = Adapter->CurBssParams.ssid.SsidLength + 1;

	dwrq->flags = 1;		/* active */

	LEAVE();
	return WLAN_STATUS_SUCCESS;
}

#ifdef	linux
/** 
 *  @brief Get version 
 *   
 *  @param adapter              A pointer to wlan_adapter structure
 *  @param version		A pointer to version buffer
 *  @param maxlen		max length of version buffer
 *  @return 	   		NA
 */
void get_version(wlan_adapter *adapter, char *version, int maxlen)
{
	union {
		u32	l;
		u8	c[4];
	} ver;
	char	fwver[32];

	ver.l = adapter->FWReleaseNumber;
	if(ver.c[3] == 0)
		sprintf(fwver, "%u.%u.%u",
			ver.c[2], ver.c[1], ver.c[0]);
	else
		sprintf(fwver, "%u.%u.%u.p%u",
			ver.c[2], ver.c[1], ver.c[0],ver.c[3]);

	snprintf(version, maxlen, driver_version, fwver);
}

#endif //linux

